###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Drop `db_dblock` table

Revision ID: django_0012
Revises: django_0011

"""

from alembic import op

revision = 'django_0012'
down_revision = 'django_0011'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    op.drop_table('db_dblock')


def downgrade():
    """Migrations for the downgrade."""
    raise NotImplementedError('Downgrade of django_0012.')



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Storage implementation using Sqlite database and disk-objectstore container."""

from __future__ import annotations

from functools import cached_property, lru_cache
from pathlib import Path
from shutil import rmtree
from typing import TYPE_CHECKING, Optional
from uuid import uuid4

from disk_objectstore import Container, backup_utils
from pydantic import BaseModel, Field, field_validator
from sqlalchemy import insert
from sqlalchemy.orm import scoped_session, sessionmaker

from aiida.common import exceptions
from aiida.common.log import AIIDA_LOGGER
from aiida.manage import Profile
from aiida.manage.configuration.settings import AIIDA_CONFIG_FOLDER
from aiida.orm.implementation import BackendEntity
from aiida.storage.psql_dos.models.settings import DbSetting
from aiida.storage.sqlite_zip import models, orm
from aiida.storage.sqlite_zip.migrator import get_schema_version_head
from aiida.storage.sqlite_zip.utils import create_sqla_engine

from ..psql_dos import PsqlDosBackend
from ..psql_dos.migrator import REPOSITORY_UUID_KEY, PsqlDosMigrator

if TYPE_CHECKING:
    from aiida.orm.entities import EntityTypes
    from aiida.repository.backend import DiskObjectStoreRepositoryBackend

__all__ = ('SqliteDosStorage',)

LOGGER = AIIDA_LOGGER.getChild(__file__)
FILENAME_DATABASE = 'database.sqlite'
FILENAME_CONTAINER = 'container'


class SqliteDosMigrator(PsqlDosMigrator):
    """Storage implementation using Sqlite database and disk-objectstore container.

    This storage backend is not recommended for use in production. The sqlite database is not the most performant and it
    does not support all the ``QueryBuilder`` functionality that is supported by the ``core.psql_dos`` storage backend.
    This storage is ideally suited for use cases that want to test or demo AiiDA as it requires no server but just a
    folder on the local filesystem.
    """

    def __init__(self, profile: Profile) -> None:
        filepath_database = Path(profile.storage_config['filepath']) / FILENAME_DATABASE
        filepath_database.touch()

        self.profile = profile
        self._engine = create_sqla_engine(filepath_database)
        self._connection = None

    def get_container(self) -> Container:
        """Return the disk-object store container.

        :returns: The disk-object store container configured for the repository path of the current profile.
        """
        filepath_container = Path(self.profile.storage_config['filepath']) / FILENAME_CONTAINER
        return Container(str(filepath_container))

    def initialise_database(self) -> None:
        """Initialise the database.

        This assumes that the database has no schema whatsoever and so the initial schema is created directly from the
        models at the current head version without migrating through all of them one by one.
        """
        models.SqliteBase.metadata.create_all(self._engine)

        repository_uuid = self.get_repository_uuid()

        # Create a "sync" between the database and repository, by saving its UUID in the settings table
        # this allows us to validate inconsistencies between the two
        self.connection.execute(
            insert(DbSetting).values(key=REPOSITORY_UUID_KEY, val=repository_uuid, description='Repository UUID')
        )

        # finally, generate the version table, "stamping" it with the most recent revision
        with self._migration_context() as context:
            context.stamp(context.script, 'main@head')  # type: ignore[arg-type]
            self.connection.commit()


class SqliteDosStorage(PsqlDosBackend):
    """A lightweight storage that is easy to install.

    This backend implementation uses an SQLite database and a disk-objectstore container as the file repository. As
    such, this storage plugin does not require any services, making it easy to install and use on most systems.
    """

    migrator = SqliteDosMigrator

    class Model(BaseModel, defer_build=True):
        """Model describing required information to configure an instance of the storage."""

        filepath: str = Field(
            title='Directory of the backend',
            description='Filepath of the directory in which to store data for this backend.',
            default_factory=lambda: str(AIIDA_CONFIG_FOLDER / 'repository' / f'sqlite_dos_{uuid4().hex}'),
        )

        @field_validator('filepath')
        @classmethod
        def filepath_is_absolute(cls, value: str) -> str:
            """Return the resolved and absolute filepath."""
            return str(Path(value).resolve().absolute())

    @property
    def filepath_root(self) -> Path:
        return Path(self.profile.storage_config['filepath'])

    @property
    def filepath_container(self) -> Path:
        return self.filepath_root / FILENAME_CONTAINER

    @property
    def filepath_database(self) -> Path:
        return self.filepath_root / FILENAME_DATABASE

    @classmethod
    def initialise(cls, profile: Profile, reset: bool = False) -> bool:
        filepath = Path(profile.storage_config['filepath'])

        try:
            filepath.mkdir(parents=True, exist_ok=True)
        except FileExistsError as exception:
            raise ValueError(
                f'`{filepath}` is a file and cannot be used for instance of `SqliteDosStorage`.'
            ) from exception

        if list(filepath.iterdir()):
            raise ValueError(
                f'`{filepath}` already exists but is not empty and cannot be used for instance of `SqliteDosStorage`.'
            )

        return super().initialise(profile, reset)

    def __str__(self) -> str:
        state = 'closed' if self.is_closed else 'open'
        return f'SqliteDosStorage[{self.filepath_root}]: {state},'

    def _initialise_session(self):
        """Initialise the SQLAlchemy session factory.

        Only one session factory is ever associated with a given class instance,
        i.e. once the instance is closed, it cannot be reopened.

        The session factory, returns a session that is bound to the current thread.
        Multi-thread support is currently required by the REST API.
        Although, in the future, we may want to move the multi-thread handling to higher in the AiiDA stack.
        """
        engine = create_sqla_engine(self.filepath_database)
        self._session_factory = scoped_session(sessionmaker(bind=engine, future=True, expire_on_commit=True))

    def delete(self) -> None:  # type: ignore[override]
        """Delete the storage and all the data."""
        if self.filepath_root.exists():
            rmtree(self.filepath_root)
            LOGGER.report(f'Deleted storage directory at `{self.filepath_root}`.')

    def get_container(self) -> 'Container':
        return Container(str(self.filepath_container))

    def get_repository(self) -> 'DiskObjectStoreRepositoryBackend':
        from aiida.repository.backend import DiskObjectStoreRepositoryBackend

        return DiskObjectStoreRepositoryBackend(container=self.get_container())

    @classmethod
    def version_head(cls) -> str:
        return get_schema_version_head()

    @classmethod
    def version_profile(cls, profile: Profile) -> str | None:
        return get_schema_version_head()

    def query(self) -> orm.SqliteQueryBuilder:
        return orm.SqliteQueryBuilder(self)

    def get_backend_entity(self, model) -> BackendEntity:
        """Return the backend entity that corresponds to the given Model instance."""
        return orm.get_backend_entity(model, self)

    @cached_property
    def authinfos(self):
        return orm.SqliteAuthInfoCollection(self)

    @cached_property
    def comments(self):
        return orm.SqliteCommentCollection(self)

    @cached_property
    def computers(self):
        return orm.SqliteComputerCollection(self)

    @cached_property
    def groups(self):
        return orm.SqliteGroupCollection(self)

    @cached_property
    def logs(self):
        return orm.SqliteLogCollection(self)

    @cached_property
    def nodes(self):
        return orm.SqliteNodeCollection(self)

    @cached_property
    def users(self):
        return orm.SqliteUserCollection(self)

    @staticmethod
    @lru_cache(maxsize=18)
    def _get_mapper_from_entity(entity_type: 'EntityTypes', with_pk: bool):
        """Return the Sqlalchemy mapper and fields corresponding to the given entity.

        :param with_pk: if True, the fields returned will include the primary key
        """
        from sqlalchemy import inspect

        from ..sqlite_zip.models import MAP_ENTITY_TYPE_TO_MODEL

        model = MAP_ENTITY_TYPE_TO_MODEL[entity_type]
        mapper = inspect(model).mapper  # type: ignore[union-attr]
        keys = {key for key, col in mapper.c.items() if with_pk or col not in mapper.primary_key}
        return mapper, keys

    def _backup(
        self,
        dest: str,
        keep: Optional[int] = None,
    ):
        """Create a backup of the storage.

        :param dest: Path to where the backup will be created. Can be a path on the local file system, or a path on a
            remote that can be accessed over SSH in the form ``<user>@<host>:<path>``.
        :param keep: The maximum number of backups to keep. If the number of copies exceeds this number, the oldest
            backups are removed.
        """
        try:
            backup_manager = backup_utils.BackupManager(dest, keep=keep)
            backup_manager.backup_auto_folders(lambda path, prev: self._backup_storage(backup_manager, path, prev))
        except backup_utils.BackupError as exc:
            raise exceptions.StorageBackupError(*exc.args) from exc

    def _backup_storage(
        self,
        manager: backup_utils.BackupManager,
        path: Path,
        prev_backup: Path | None = None,
    ) -> None:
        """Create a backup of the sqlite database and disk-objectstore to the provided path.

        :param manager: BackupManager from backup_utils containing utilities such as for calling the rsync.
        :param path: Path to where the backup will be created.
        :param prev_backup: Path to the previous backup. Rsync calls will be hard-linked to this path, making the backup
            incremental and efficient.
        """
        LOGGER.report('Running storage maintenance')
        self.maintain(full=False, compress=False)

        LOGGER.report('Backing up disk-objectstore container')
        manager.call_rsync(self.filepath_container, path, link_dest=prev_backup, dest_trailing_slash=True)

        LOGGER.report('Backing up sqlite database')
        manager.call_rsync(self.filepath_database, path, link_dest=prev_backup, dest_trailing_slash=True)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Drop the hashes for all ``CalcJobNode`` instances.

The computed hash erroneously included the hash of the file repository. This was present as of v2.0 and so all nodes
created with versions since then will have incorrect hashes.

Revision ID: main_0002
Revises: main_0001
Create Date: 2023-05-05

"""

from alembic import op

from aiida.storage.psql_dos.migrations.utils.integrity import drop_hashes

revision = 'main_0002'
down_revision = 'main_0001'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    drop_hashes(
        op.get_bind(), hash_extra_key='_aiida_hash', entry_point_string='aiida.node:process.calculation.calcjob'
    )


def downgrade():
    """Migrations for the downgrade."""
    drop_hashes(
        op.get_bind(), hash_extra_key='_aiida_hash', entry_point_string='aiida.node:process.calculation.calcjob'
    )



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Collection of ``pytest`` fixtures that are intended for use in plugin packages.

To use these fixtures, simply create a ``conftest.py`` in the tests folder and add the following line:

    pytest_plugins = ['aiida.manage.tests.pytest_fixtures']

This will make all the fixtures in this file available and ready for use. Simply use them as you would any other
``pytest`` fixture.
"""

from __future__ import annotations

import asyncio
import contextlib
import inspect
import io
import os
import pathlib
import shutil
import tempfile
import time
import typing as t
import uuid
import warnings

import plumpy
import pytest
import wrapt
from importlib_metadata import EntryPoint, EntryPoints

from aiida import plugins
from aiida.common.exceptions import NotExistent
from aiida.common.lang import type_check
from aiida.common.log import AIIDA_LOGGER
from aiida.common.warnings import warn_deprecation
from aiida.engine import Process, ProcessBuilder, submit
from aiida.engine.daemon.client import DaemonClient, DaemonNotRunningException, DaemonTimeoutException
from aiida.manage import Profile, get_manager, get_profile
from aiida.manage.manager import Manager
from aiida.orm import Computer, ProcessNode, User

if t.TYPE_CHECKING:
    from aiida.manage.configuration.config import Config


warn_deprecation(
    'The module `aiida.manage.tests.pytest_fixtures` is deprecated, please use `aiida.tools.pytest_fixtures` instead.',
    version=3,
)


def recursive_merge(left: dict[t.Any, t.Any], right: dict[t.Any, t.Any]) -> None:
    """Recursively merge the ``right`` dictionary into the ``left`` dictionary.

    :param left: Base dictionary.
    :param right: Dictionary to recurisvely merge on top of ``left`` dictionary.
    """
    for key, value in right.items():
        if key in left and isinstance(left[key], dict) and isinstance(value, dict):
            recursive_merge(left[key], value)
        else:
            left[key] = value


@pytest.fixture(scope='function')
def aiida_caplog(caplog):
    """A copy of pytest's caplog fixture, which allows ``AIIDA_LOGGER`` to propagate."""
    propogate = AIIDA_LOGGER.propagate
    AIIDA_LOGGER.propagate = True
    yield caplog
    AIIDA_LOGGER.propagate = propogate


@pytest.fixture(scope='session')
def postgres_cluster():
    """Create a temporary and isolated PostgreSQL cluster using ``pgtest`` and cleanup after the yield.

    :param database_name: Name of the database.
    :param database_username: Username to use for authentication.
    :param database_password: Password to use for authentication.
    :returns: Dictionary with parameters to connect to the PostgreSQL cluster.
    """
    from pgtest.pgtest import PGTest

    def create_database(
        database_name: str | None = None, database_username: str | None = None, database_password: str | None = None
    ) -> t.Generator[dict[str, str], None, None]:
        from aiida.manage.external.postgres import Postgres

        postgres_config = {
            'database_engine': 'postgresql_psycopg2',
            'database_name': database_name or str(uuid.uuid4()),
            'database_username': database_username or 'guest',
            'database_password': database_password or 'guest',
        }

        postgres = Postgres(interactive=False, quiet=True, dbinfo=cluster.dsn)
        if not postgres.dbuser_exists(postgres_config['database_username']):
            postgres.create_dbuser(
                postgres_config['database_username'], postgres_config['database_password'], 'CREATEDB'
            )
        postgres.create_db(postgres_config['database_username'], postgres_config['database_name'])

        postgres_config['database_hostname'] = postgres.host_for_psycopg2
        postgres_config['database_port'] = postgres.port_for_psycopg2

        return postgres_config

    cluster = None
    try:
        cluster = PGTest()
        cluster.create_database = create_database
        yield cluster
    finally:
        if cluster is not None:
            cluster.close()


@pytest.fixture(scope='session')
def aiida_test_profile() -> str | None:
    """Return the name of the AiiDA test profile if defined.

    The name is taken from the ``AIIDA_TEST_PROFILE`` environment variable.

    :returns: The name of the profile to you for the test session or ``None`` if not defined.
    """
    return os.environ.get('AIIDA_TEST_PROFILE', None)


@pytest.fixture(scope='session')
def aiida_manager() -> Manager:
    """Return the global instance of the :class:`~aiida.manage.manager.Manager`.

    :returns: The global manager instance.
    """
    return get_manager()


@pytest.fixture(scope='session')
def aiida_instance(
    tmp_path_factory: pytest.tmpdir.TempPathFactory,
    aiida_manager: Manager,
    aiida_test_profile: str | None,
) -> t.Generator[Config, None, None]:
    """Return the :class:`~aiida.manage.configuration.config.Config` instance that is used for the test session.

    If an existing test profile is defined through the ``aiida_test_profile`` fixture, the configuration of the actual
    AiiDA instance is loaded and returned. If no test profile is defined, a completely independent and temporary AiiDA
    instance is generated in a temporary directory with a clean `.aiida` folder and basic configuration file. The
    currently loaded configuration and profile are stored in memory and are automatically restored at the end of the
    test session. The temporary instance is automatically deleted.

    :return: The configuration the AiiDA instance loaded for this test session.
    """
    from aiida.manage import configuration
    from aiida.manage.configuration import settings

    if aiida_test_profile:
        yield configuration.get_config()

    else:
        reset = False

        if configuration.CONFIG is not None:
            reset = True
            current_config = configuration.CONFIG
            current_config_path = current_config.dirpath
            current_profile = configuration.get_profile()
            current_path_variable = os.environ.get(settings.DEFAULT_AIIDA_PATH_VARIABLE, None)

        dirpath_config = tmp_path_factory.mktemp('config')
        os.environ[settings.DEFAULT_AIIDA_PATH_VARIABLE] = str(dirpath_config)
        settings.AIIDA_CONFIG_FOLDER = dirpath_config
        settings.set_configuration_directory()
        configuration.CONFIG = configuration.load_config(create=True)

        try:
            yield configuration.CONFIG
        finally:
            if reset:
                if current_path_variable is None:
                    os.environ.pop(settings.DEFAULT_AIIDA_PATH_VARIABLE, None)
                else:
                    os.environ[settings.DEFAULT_AIIDA_PATH_VARIABLE] = current_path_variable

                settings.AIIDA_CONFIG_FOLDER = current_config_path
                configuration.CONFIG = current_config
                if current_profile:
                    aiida_manager.load_profile(current_profile.name, allow_switch=True)


@pytest.fixture(scope='session')
def config_psql_dos(
    tmp_path_factory: pytest.tmpdir.TempPathFactory,
    postgres_cluster: dict[str, str],
) -> t.Callable[[dict[str, t.Any] | None], dict[str, t.Any]]:
    """Return a profile configuration for the :class:`~aiida.storage.psql_dos.backend.PsqlDosBackend`."""

    def factory(custom_configuration: dict[str, t.Any] | None = None) -> dict[str, t.Any]:
        """Return a profile configuration for the :class:`~aiida.storage.psql_dos.backend.PsqlDosBackend`.

        :param custom_configuration: Custom configuration to override default profile configuration.
        :returns: The profile configuration.
        """
        custom_configuration = custom_configuration or {}
        custom_configuration.setdefault('storage', {}).setdefault('config', {})
        custom_storage_config = custom_configuration['storage']['config']

        storage_config = postgres_cluster.create_database(
            database_name=custom_storage_config.get('database_name'),
            database_username=custom_storage_config.get('database_username'),
            database_password=custom_storage_config.get('database_password'),
        )
        recursive_merge(custom_storage_config, storage_config)

        configuration = {
            'storage': {
                'backend': 'core.psql_dos',
                'config': {
                    'repository_uri': f'file://{tmp_path_factory.mktemp("repository")}',
                },
            }
        }
        recursive_merge(configuration, custom_configuration)
        return configuration

    return factory


def clear_profile():
    """Clear the currently loaded profile.

    This ensures that the contents of the profile are reset as well as the ``Manager``, which may hold references to
    data that will be destroyed. The daemon will also be stopped if it was running.
    """
    from aiida.engine.daemon.client import get_daemon_client

    daemon_client = get_daemon_client()

    if daemon_client.is_daemon_running:
        daemon_client.stop_daemon(wait=True)

    manager = get_manager()
    manager.get_profile_storage()._clear()
    manager.reset_broker()
    manager.reset_runner()

    User(get_manager().get_profile().default_user_email).store()


@pytest.fixture(scope='session')
def aiida_profile_factory(
    aiida_instance: Config,
    aiida_manager: Manager,
) -> t.Callable[[dict[str, t.Any]], Profile]:
    """Create a temporary profile, add it to the config of the loaded AiiDA instance and load the profile.

    The default configuration is complete except for the configuration of the storage, which should be provided through
    the ``custom_configuration`` argument. The storage will be fully reset and initalised, destroying all data that it
    contains and recreate the default user, making the profile ready for use.
    """

    def factory(custom_configuration: dict[str, t.Any] | None = None) -> Profile:
        """Create an isolated AiiDA instance with a temporary and fully loaded profile.

        :param custom_configuration: Custom configuration to override default profile configuration.
        :returns: The constructed profile.
        """
        config = aiida_instance
        configuration = {
            'default_user_email': 'test@aiida.local',
            'storage': {},
            'process_control': {
                'backend': 'rabbitmq',
                'config': {
                    'broker_protocol': 'amqp',
                    'broker_username': 'guest',
                    'broker_password': 'guest',
                    'broker_host': '127.0.0.1',
                    'broker_port': 5672,
                    'broker_virtual_host': '',
                },
            },
            'options': {
                'warnings.development_version': False,
                'warnings.rabbitmq_version': False,
            },
        }
        recursive_merge(configuration, custom_configuration or {})
        configuration['test_profile'] = True

        with contextlib.redirect_stdout(io.StringIO()):
            profile_name = str(uuid.uuid4())
            profile = Profile(profile_name, configuration)
            profile.storage_cls.initialise(profile, reset=True)

            config.add_profile(profile)
            config.set_default_profile(profile_name)
            config.store()

            aiida_manager.load_profile(profile_name, allow_switch=True)

            User(profile.default_user_email).store()

        # Add the ``clear_profile`` method, such that users can empty the storage through the ``Profile`` instance that
        # is returned by this fixture. This functionality is added for backwards-compatibility as before the fixture
        # used to return an instance of the :class:`~aiida.manage.tests.main.TestManager` which provided this method
        # that was often used.
        setattr(profile, 'clear_profile', clear_profile)

        return profile

    return factory


@pytest.fixture(scope='session', autouse=True)
def aiida_profile(
    aiida_manager: Manager,
    aiida_test_profile: str | None,
    aiida_profile_factory: t.Callable[[dict[str, t.Any] | None], Profile],
    config_psql_dos: t.Callable[[dict[str, t.Any] | None], dict[str, t.Any]],
) -> t.Generator[Profile, None, None]:
    """Return a loaded AiiDA test profile.

    If a test profile has been declared, as returned by the ``aiida_test_profile`` fixture, that is loaded and yielded.
    Otherwise, a temporary and fully isolated AiiDA instance is created, complete with a loaded test profile, that are
    all automatically cleaned up at the end of the test session. The storage backend used for the profile is
    :class:`~aiida.storage.psql_dos.backend.PsqlDosBackend`.
    """
    if aiida_test_profile is not None:
        aiida_manager.load_profile(aiida_test_profile)
        profile = get_profile()

        if profile is None:
            raise RuntimeError(f'could not load the `{aiida_test_profile}` test profile.')

        if not profile.is_test_profile:
            raise RuntimeError(f'specified test profile `{aiida_test_profile}` is not a test profile.')

        # Add the ``clear_profile`` method. See ``aiida_profile_factory`` for the reasoning. Note that since it is added
        # there, this only needs to be added here, for an existing test profile, because the temporarily created profile
        # will have it added by the ``aiida_profile_factory`` fixture itself.
        setattr(profile, 'clear_profile', clear_profile)
    else:
        profile = aiida_profile_factory(config_psql_dos({}))

    yield profile


@pytest.fixture(scope='function')
def aiida_profile_clean(aiida_profile):
    """Provide an AiiDA test profile, with the storage reset at test function setup."""
    aiida_profile.clear_profile()
    yield aiida_profile


@pytest.fixture(scope='class')
def aiida_profile_clean_class(aiida_profile):
    """Provide an AiiDA test profile, with the storage reset at test class setup."""
    aiida_profile.clear_profile()
    yield aiida_profile


@pytest.fixture(scope='function')
def clear_database(clear_database_after_test):
    """Alias for 'clear_database_after_test'.

    Clears the database after each test. Use of the explicit
    'clear_database_after_test' is preferred.
    """


@pytest.fixture(scope='function')
def clear_database_after_test(aiida_profile):
    """Clear the database after the test."""
    warn_deprecation('the clear_database_after_test fixture is deprecated, use aiida_profile_clean instead', version=3)
    yield aiida_profile
    aiida_profile.clear_profile()


@pytest.fixture(scope='function')
def clear_database_before_test(aiida_profile):
    """Clear the database before the test."""
    warn_deprecation('the clear_database_before_test fixture deprecated, use aiida_profile_clean instead', version=3)
    aiida_profile.clear_profile()
    yield aiida_profile


@pytest.fixture(scope='class')
def clear_database_before_test_class(aiida_profile):
    """Clear the database before a test class."""
    warn_deprecation(
        'the clear_database_before_test_class is deprecated, use aiida_profile_clean_class instead', version=3
    )
    aiida_profile.clear_profile()
    yield


@pytest.fixture(scope='function')
def temporary_event_loop():
    """Create a temporary loop for independent test case"""
    current = asyncio.get_event_loop()
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    try:
        yield
    finally:
        loop.close()
        asyncio.set_event_loop(current)


@pytest.fixture(scope='function')
def temp_dir():
    """Get a temporary directory.

    E.g. to use as the working directory of an AiiDA computer.

    :return: The path to the directory
    :rtype: str

    """
    warn_deprecation('This fixture is deprecated, use the `tmp_path` fixture of `pytest` instead.', version=3)
    try:
        dirpath = tempfile.mkdtemp()
        yield dirpath
    finally:
        # after the test function has completed, remove the directory again
        shutil.rmtree(dirpath)


@pytest.fixture
def aiida_local_code_factory(aiida_localhost):
    """Get an AiiDA code on localhost.

    Searches in the PATH for a given executable and creates an AiiDA code with provided entry point.

    Usage::

      def test_1(aiida_local_code_factory):
          code = aiida_local_code_factory('quantumespresso.pw', '/usr/bin/pw.x')
          # use code for testing ...

    :return: A function get_code(entry_point, executable) that returns the `Code` node.
    :rtype: object
    """

    def get_code(entry_point, executable, computer=aiida_localhost, label=None, **kwargs):
        """Get local code.

        Sets up code for given entry point on given computer.

        :param entry_point: Entry point of calculation plugin
        :param executable: name of executable; will be searched for in local system PATH.
        :param computer: (local) AiiDA computer
        :param label: Define the label of the code. By default the ``executable`` is taken. This can be useful if
            multiple codes need to be created in a test which require unique labels.
        :param kwargs: Additional keyword arguments that are passed to the code's constructor.
        :return: the `Code` either retrieved from the database or created if it did not yet exist.
        :rtype: :py:class:`~aiida.orm.Code`
        """
        from aiida.common import exceptions
        from aiida.orm import InstalledCode, QueryBuilder

        if label is None:
            label = executable

        builder = QueryBuilder().append(Computer, filters={'uuid': computer.uuid}, tag='computer')
        builder.append(
            InstalledCode, filters={'label': label, 'attributes.input_plugin': entry_point}, with_computer='computer'
        )

        try:
            code = builder.one()[0]
        except (exceptions.MultipleObjectsError, exceptions.NotExistent):
            code = None
        else:
            return code

        executable_path = shutil.which(executable)
        if not executable_path:
            raise ValueError(f'The executable "{executable}" was not found in the $PATH.')

        code = InstalledCode(
            label=label,
            description=label,
            default_calc_job_plugin=entry_point,
            computer=computer,
            filepath_executable=executable_path,
            **kwargs,
        )

        return code.store()

    return get_code


@pytest.fixture(scope='session')
def ssh_key(tmp_path_factory) -> t.Generator[pathlib.Path, None, None]:
    """Generate a temporary SSH key pair for the test session and return the filepath of the private key.

    The filepath of the public key is the same as the private key, but it adds the ``.pub`` file extension.
    """
    from cryptography.hazmat.backends import default_backend as crypto_default_backend
    from cryptography.hazmat.primitives import serialization as crypto_serialization
    from cryptography.hazmat.primitives.asymmetric import rsa

    key = rsa.generate_private_key(
        backend=crypto_default_backend(),
        public_exponent=65537,
        key_size=2048,
    )

    private_key = key.private_bytes(
        crypto_serialization.Encoding.PEM,
        crypto_serialization.PrivateFormat.PKCS8,
        crypto_serialization.NoEncryption(),
    )

    public_key = key.public_key().public_bytes(
        crypto_serialization.Encoding.OpenSSH,
        crypto_serialization.PublicFormat.OpenSSH,
    )

    dirpath = tmp_path_factory.mktemp('keys')
    filename = uuid.uuid4().hex
    filepath_private_key = dirpath / filename
    filepath_public_key = dirpath / f'{filename}.pub'

    filepath_private_key.write_bytes(private_key)
    filepath_public_key.write_bytes(public_key)

    try:
        yield filepath_private_key
    finally:
        filepath_private_key.unlink(missing_ok=True)
        filepath_public_key.unlink(missing_ok=True)


@pytest.fixture
def aiida_computer(tmp_path) -> t.Callable[[], Computer]:
    """Factory to return a :class:`aiida.orm.computers.Computer` instance."""

    def factory(
        label: t.Optional[str] = None,
        minimum_job_poll_interval: int = 0,
        default_mpiprocs_per_machine: int = 1,
        configuration_kwargs: t.Optional[dict[t.Any, t.Any]] = None,
        **kwargs,
    ) -> Computer:
        """Return a :class:`aiida.orm.computers.Computer` instance.

        The database is queried for an existing computer with the given label. If it exists, it means it was probably
        created by this fixture in a previous call and it is simply returned. Otherwise a new instance is created.
        Note that the computer is not explicitly configured, unless ``configure_kwargs`` are specified.

        :param label: The computer label. If not specified, a random UUID4 is used.
        :param minimum_job_poll_interval: The default minimum job poll interval to set.
        :param configuration_kwargs: Optional keyword arguments that, if defined, are used to configure the computer
            by calling :meth:`aiida.orm.computers.Computer.configure`.
        :param kwargs: Optional keyword arguments that are passed to the :class:`aiida.orm.computers.Computer`
            constructor if a computer with the given label doesn't already exist.
        :return: A stored computer instance.
        """
        label = label or str(uuid.uuid4())

        try:
            computer = Computer.collection.get(label=label)
        except NotExistent:
            computer = Computer(
                label=label,
                description=kwargs.pop('description', 'computer created by `aiida_computer` fixture'),
                hostname=kwargs.pop('hostname', 'localhost'),
                workdir=kwargs.pop('workdir', str(tmp_path)),
                transport_type=kwargs.pop('transport_type', 'core.local'),
                scheduler_type=kwargs.pop('scheduler_type', 'core.direct'),
            )
            computer.store()
            computer.set_minimum_job_poll_interval(minimum_job_poll_interval)
            computer.set_default_mpiprocs_per_machine(default_mpiprocs_per_machine)

        if configuration_kwargs:
            computer.configure(**configuration_kwargs)

        return computer

    return factory


@pytest.fixture
def aiida_computer_local(aiida_computer) -> t.Callable[[], Computer]:
    """Factory to return a :class:`aiida.orm.computers.Computer` instance with ``core.local`` transport."""

    def factory(label: t.Optional[str] = None, configure: bool = True) -> Computer:
        """Return a :class:`aiida.orm.computers.Computer` instance representing localhost with ``core.local`` transport.

        The database is queried for an existing computer with the given label. If it exists, it is returned, otherwise a
        new instance is created.

        :param label: The computer label. If not specified, a random UUID4 is used.
        :param configure: Boolean, if ``True``, ensures the computer is configured, otherwise the computer is returned
            as is. Note that if a computer with the given label already exists and it was configured before, the
            computer will not be "un-"configured. If an unconfigured computer is absolutely required, make sure to first
            delete the existing computer or specify another label.
        :return: A stored computer instance.
        """
        computer = aiida_computer(label=label, hostname='localhost', transport_type='core.local')

        if configure:
            computer.configure()

        return computer

    return factory


@pytest.fixture
def aiida_computer_ssh(aiida_computer, ssh_key) -> t.Callable[[], Computer]:
    """Factory to return a :class:`aiida.orm.computers.Computer` instance with ``core.ssh`` transport."""

    def factory(label: t.Optional[str] = None, configure: bool = True) -> Computer:
        """Return a :class:`aiida.orm.computers.Computer` instance representing localhost with ``core.ssh`` transport.

        The database is queried for an existing computer with the given label. If it exists, it is returned, otherwise a
        new instance is created.

        If ``configure=True``, an SSH key pair is automatically added to the ``.ssh`` folder of the user, allowing an
        actual SSH connection to be made to the localhost.

        :param label: The computer label. If not specified, a random UUID4 is used.
        :param configure: Boolean, if ``True``, ensures the computer is configured, otherwise the computer is returned
            as is. Note that if a computer with the given label already exists and it was configured before, the
            computer will not be "un-"configured. If an unconfigured computer is absolutely required, make sure to first
            delete the existing computer or specify another label.
        :return: A stored computer instance.
        """
        computer = aiida_computer(label=label, hostname='localhost', transport_type='core.ssh')

        if configure:
            computer.configure(
                key_filename=str(ssh_key),
                key_policy='AutoAddPolicy',
            )

        return computer

    return factory


@pytest.fixture(scope='function')
def aiida_localhost(aiida_computer_local) -> Computer:
    """Return a :class:`aiida.orm.computers.Computer` instance representing localhost with ``core.local`` transport.

    Usage::

        def test(aiida_localhost):
            assert aiida_localhost.transport_type == 'core.local'

    :return: The computer.
    """
    return aiida_computer_local(label='localhost')


@pytest.fixture(scope='session')
def daemon_client(aiida_profile):
    """Return a daemon client for the configured test profile for the test session.

    The daemon will be automatically stopped at the end of the test session.
    """
    daemon_client = DaemonClient(aiida_profile)

    try:
        yield daemon_client
    finally:
        try:
            daemon_client.stop_daemon(wait=True)
        except DaemonNotRunningException:
            pass
        # Give an additional grace period by manually waiting for the daemon to be stopped. In certain unit test
        # scenarios, the built in wait time in ``daemon_client.stop_daemon`` is not sufficient and even though the
        # daemon is stopped, ``daemon_client.is_daemon_running`` will return false for a little bit longer.
        daemon_client._await_condition(
            lambda: not daemon_client.is_daemon_running,
            DaemonTimeoutException('The daemon failed to stop.'),
        )


@pytest.fixture()
def started_daemon_client(daemon_client):
    """Ensure that the daemon is running for the test profile and return the associated client."""
    if not daemon_client.is_daemon_running:
        daemon_client.start_daemon()
        assert daemon_client.is_daemon_running

    yield daemon_client


@pytest.fixture()
def stopped_daemon_client(daemon_client):
    """Ensure that the daemon is not running for the test profile and return the associated client."""
    if daemon_client.is_daemon_running:
        daemon_client.stop_daemon(wait=True)
        # Give an additional grace period by manually waiting for the daemon to be stopped. In certain unit test
        # scenarios, the built in wait time in ``daemon_client.stop_daemon`` is not sufficient and even though the
        # daemon is stopped, ``daemon_client.is_daemon_running`` will return false for a little bit longer.
        daemon_client._await_condition(
            lambda: not daemon_client.is_daemon_running,
            DaemonTimeoutException('The daemon failed to stop.'),
        )

    yield daemon_client


@pytest.fixture
def submit_and_await(started_daemon_client):
    """Submit a process and wait for it to achieve the given state."""

    def _factory(
        submittable: Process | ProcessBuilder | ProcessNode,
        state: plumpy.ProcessState = plumpy.ProcessState.FINISHED,
        timeout: int = 20,
        **kwargs,
    ):
        """Submit a process and wait for it to achieve the given state.

        :param submittable: A process, a process builder or a process node. If it is a process or builder, it is
            submitted first before awaiting the desired state.
        :param state: The process state to wait for, by default it waits for the submittable to be ``FINISHED``.
        :param timeout: The time to wait for the process to achieve the state.
        :param kwargs: If the ``submittable`` is a process class, it is instantiated with the ``kwargs`` as inputs.
        :raises RuntimeError: If the process fails to achieve the specified state before the timeout expires.
        """
        if inspect.isclass(submittable) and issubclass(submittable, Process):
            node = submit(submittable, **kwargs)
        elif isinstance(submittable, ProcessBuilder):
            node = submit(submittable)
        elif isinstance(submittable, ProcessNode):
            node = submittable
        else:
            raise ValueError(f'type of submittable `{type(submittable)}` is not supported.')

        start_time = time.time()

        while node.process_state is not state:
            if node.is_excepted:
                raise RuntimeError(f'The process excepted: {node.exception}')

            if time.time() - start_time >= timeout:
                daemon_log_file = pathlib.Path(started_daemon_client.daemon_log_file).read_text(encoding='utf-8')
                daemon_status = 'running' if started_daemon_client.is_daemon_running else 'stopped'
                raise RuntimeError(
                    f'Timed out waiting for process with state `{node.process_state}` to enter state `{state}`.\n'
                    f'Daemon <{started_daemon_client.profile.name}|{daemon_status}> log file content: \n'
                    f'{daemon_log_file}'
                )

        return node

    return _factory


@wrapt.decorator
def suppress_deprecations(wrapped, _, args, kwargs):
    """Decorator that suppresses all ``DeprecationWarning``."""
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=DeprecationWarning)
        return wrapped(*args, **kwargs)


class EntryPointManager:
    """Manager to temporarily add or remove entry points."""

    def __init__(self, entry_points: EntryPoints):
        self.entry_points = entry_points

    def eps(self) -> EntryPoints:
        return self.entry_points

    def eps_select(self, group, name=None) -> EntryPoints:
        if name is None:
            return self.eps().select(group=group)
        return self.eps().select(group=group, name=name)

    @staticmethod
    def _validate_entry_point(entry_point_string: str | None, group: str | None, name: str | None) -> tuple[str, str]:
        """Validate the definition of the entry point.

        :param entry_point_string: Fully qualified entry point string.
        :param name: Entry point name.
        :param group: Entry point group.
        :returns: The entry point group and name.
        :raises TypeError: If `entry_point_string`, `group` or `name` are not a string, when defined.
        :raises ValueError: If `entry_point_string` is not defined, nor a `group` and `name`.
        :raises ValueError: If `entry_point_string` is not a complete entry point string with group and name.
        """
        if entry_point_string is not None:
            try:
                group, name = plugins.entry_point.parse_entry_point_string(entry_point_string)
            except TypeError as exception:
                raise TypeError('`entry_point_string` should be a string when defined.') from exception
            except ValueError as exception:
                raise ValueError('invalid `entry_point_string` format, should `group:name`.') from exception

        if name is None or group is None:
            raise ValueError('neither `entry_point_string` is defined, nor `name` and `group`.')

        type_check(group, str)
        type_check(name, str)

        return group, name

    def add(
        self,
        value: type | str,
        entry_point_string: str | None = None,
        *,
        name: str | None = None,
        group: str | None = None,
    ) -> None:
        """Add an entry point.

        :param value: The class or function to register as entry point. The resource needs to be importable, so it can't
            be inlined. Alternatively, the fully qualified name can be passed as a string.
        :param entry_point_string: Fully qualified entry point string.
        :param name: Entry point name.
        :param group: Entry point group.
        :returns: The entry point group and name.
        :raises TypeError: If `entry_point_string`, `group` or `name` are not a string, when defined.
        :raises ValueError: If `entry_point_string` is not defined, nor a `group` and `name`.
        :raises ValueError: If `entry_point_string` is not a complete entry point string with group and name.
        """
        if not isinstance(value, str):
            value = f'{value.__module__}:{value.__name__}'

        group, name = self._validate_entry_point(entry_point_string, group, name)
        entry_point = EntryPoint(name, value, group)
        self.entry_points = EntryPoints(self.entry_points + (entry_point,))

    def remove(
        self, entry_point_string: str | None = None, *, name: str | None = None, group: str | None = None
    ) -> None:
        """Remove an entry point.

        :param value: Entry point value, fully qualified import path name.
        :param entry_point_string: Fully qualified entry point string.
        :param name: Entry point name.
        :param group: Entry point group.
        :returns: The entry point group and name.
        :raises TypeError: If `entry_point_string`, `group` or `name` are not a string, when defined.
        :raises ValueError: If `entry_point_string` is not defined, nor a `group` and `name`.
        :raises ValueError: If `entry_point_string` is not a complete entry point string with group and name.
        """
        group, name = self._validate_entry_point(entry_point_string, group, name)
        try:
            self.entry_points[name]
        except KeyError:
            raise KeyError(f'entry point `{name}` does not exist in group `{group}`.')
        self.entry_points = EntryPoints((ep for ep in self.entry_points if not (ep.name == name and ep.group == group)))


@pytest.fixture
def entry_points(monkeypatch) -> EntryPointManager:
    """Return an instance of the ``EntryPointManager`` which allows to temporarily add or remove entry points.

    This fixture monkey patches the entry point caches returned by
    the :func:`aiida.plugins.entry_point.eps` and :func:`aiida.plugins.entry_point.eps_select` functions
    to class methods of the ``EntryPointManager`` so that we can dynamically add / remove entry points.
    Note that we do not need a deepcopy here as ``eps()`` returns an immutable ``EntryPoints`` tuple type.
    """
    epm = EntryPointManager(plugins.entry_point.eps())
    monkeypatch.setattr(plugins.entry_point, 'eps', epm.eps)
    monkeypatch.setattr(plugins.entry_point, 'eps_select', epm.eps_select)
    yield epm



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""This migration creates UUID column and populates it with distinct UUIDs

This migration corresponds to the 0024_dblog_update Django migration (only the final part).

Revision ID: ea2f50e7f615
Revises: 041a79fc615f
Create Date: 2019-01-30 19:22:50.984380
"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = 'ea2f50e7f615'
down_revision = '041a79fc615f'
branch_labels = None
depends_on = None


def upgrade():
    """Add an UUID column an populate it with unique UUIDs"""
    from aiida.common.utils import get_new_uuid
    from aiida.storage.psql_dos.migrations.utils.dblog_update import set_new_uuid

    connection = op.get_bind()

    # Create the UUID column
    op.add_column(
        'db_dblog', sa.Column('uuid', postgresql.UUID(), autoincrement=False, nullable=True, default=get_new_uuid)
    )

    # Populate the uuid column
    set_new_uuid(connection)


def downgrade():
    """Remove the UUID column"""
    op.drop_column('db_dblog', 'uuid')



"""Check that the GitHub release tag matches the package version."""

import argparse
import ast
from pathlib import Path


def get_version_from_module(content: str) -> str:
    """Get the __version__ value from a module."""
    # adapted from setuptools/config.py
    try:
        module = ast.parse(content)
    except SyntaxError as exc:
        raise OSError(f'Unable to parse module: {exc}')
    try:
        return next(
            ast.literal_eval(statement.value)
            for statement in module.body
            if isinstance(statement, ast.Assign)
            for target in statement.targets
            if isinstance(target, ast.Name) and target.id == '__version__'
        )
    except StopIteration:
        raise OSError('Unable to find __version__ in module')


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('GITHUB_REF', help='The GITHUB_REF environmental variable')
    args = parser.parse_args()
    assert args.GITHUB_REF.startswith('refs/tags/v'), f'GITHUB_REF should start with "refs/tags/v": {args.GITHUB_REF}'
    tag_version = args.GITHUB_REF[11:]
    pypi_version = get_version_from_module(Path('src/aiida/__init__.py').read_text(encoding='utf-8'))
    assert tag_version == pypi_version, f'The tag version {tag_version} != {pypi_version} specified in `pyproject.toml`'



"""Module with data plugins that represent remote resources and so effectively are symbolic links."""

# AUTO-GENERATED

# fmt: off

from .base import *
from .stash import *

__all__ = (
    'RemoteData',
    'RemoteStashData',
    'RemoteStashFolderData',
)

# fmt: on



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Utility functions to draw ASCII diagrams to the command line."""

from typing import Optional

__all__ = ('format_call_graph',)

TREE_LAST_ENTRY = '\u2514\u2500\u2500 '
TREE_MIDDLE_ENTRY = '\u251c\u2500\u2500 '
TREE_FIRST_ENTRY = TREE_MIDDLE_ENTRY


def calc_info(node, call_link_label: bool = False) -> str:
    """Return a string with the summary of the state of a CalculationNode.

    :param calc_node: The calculation node
    :param call_link_label: Include the call link label if other from the default ``CALL``.
    """
    from aiida.orm import ProcessNode, WorkChainNode

    if not isinstance(node, ProcessNode):
        raise TypeError(f'Unknown type: {type(node)}')

    process_label = node.process_label
    process_state = node.process_state.value.capitalize()
    exit_status = node.exit_status

    if call_link_label and (caller := node.caller):
        from aiida.common.links import LinkType

        call_link = next(
            triple.link_label
            for triple in caller.base.links.get_outgoing(link_type=(LinkType.CALL_CALC, LinkType.CALL_WORK)).all()
            if triple.node.pk == node.pk
        )
    else:
        call_link = None

    if call_link and call_link != 'CALL':
        string = f'{process_label}<{node.pk} | {call_link}> {process_state}'
    else:
        string = f'{process_label}<{node.pk}> {process_state}'

    if exit_status is not None:
        string += f' [{exit_status}]'

    if isinstance(node, WorkChainNode) and node.stepper_state_info:
        string += f' [{node.stepper_state_info}]'

    return string


def format_call_graph(calc_node, max_depth: Optional[int] = None, call_link_label: bool = False, info_fn=calc_info):
    """Print a tree like the POSIX tree command for the calculation call graph.

    :param calc_node: The calculation node
    :param max_depth: Maximum depth of the call graph to print
    :param call_link_label: Include the call link label if other from the default ``CALL``.
    :param info_fn: An optional function that takes the node and returns a string of information to be displayed for
        each node.
    """
    call_tree = build_call_graph(calc_node, max_depth=max_depth, call_link_label=call_link_label, info_fn=info_fn)
    return format_tree_descending(call_tree)


def build_call_graph(
    calc_node, max_depth: Optional[int] = None, call_link_label: bool = False, info_fn=calc_info
) -> str:
    """Build the call graph of a given node.

    :param calc_node: The calculation node
    :param max_depth: Maximum depth of the call graph to build. Use `None` for unlimited.
    :param call_link_label: Include the call link label if other from the default ``CALL``.
    :param info_fn: An optional function that takes the node and returns a string of information to be displayed for
        each node.
    """
    if max_depth is not None:
        if max_depth < 0:
            raise ValueError('max_depth must be >= 0')
        if max_depth == 0:
            return ''

    info_string = info_fn(calc_node, call_link_label)
    called = calc_node.called
    called.sort(key=lambda x: x.ctime)
    if called and max_depth != 1:
        if max_depth is not None:
            max_depth -= 1
        return info_string, [
            build_call_graph(c, max_depth=max_depth, call_link_label=call_link_label, info_fn=info_fn) for c in called
        ]

    return info_string


def format_tree_descending(tree, prefix='', pos=-1):
    """Format a descending tree."""
    text = []

    if isinstance(tree, tuple):
        info = tree[0]
    else:
        info = tree

    if pos == -1:
        pre = ''
    elif pos == 0:
        pre = f'{prefix}{TREE_FIRST_ENTRY}'
    elif pos == 1:
        pre = f'{prefix}{TREE_MIDDLE_ENTRY}'
    else:
        pre = f'{prefix}{TREE_LAST_ENTRY}'
    text.append(f'{pre}{info}')

    if isinstance(tree, tuple):
        _, value = tree
        num_entries = len(value)
        if pos in [-1, 2]:
            new_prefix = f'{prefix}    '
        else:
            new_prefix = f'{prefix}│   '
        for i, entry in enumerate(value):
            if i == num_entries - 1:
                pos = 2
            elif i == 0:
                pos = 0
            else:
                pos = 1
            text.append(format_tree_descending(entry, new_prefix, pos))

    return '\n'.join(text)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Clean the log records from non-Node entity records (part a).

It removes from the ``DbLog`` table, the legacy workflow records and records
that correspond to an unknown entity and places them to corresponding files.

Note this migration is similar to the sqlalchemy migration 041a79fc615f + ea2f50e7f615

Revision ID: django_0024a
Revises: django_0023

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

from aiida.storage.psql_dos.migrations.utils.dblog_update import export_and_clean_workflow_logs, set_new_uuid

revision = 'django_0024a'
down_revision = 'django_0023'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    connection = op.get_bind()

    # Clean data
    export_and_clean_workflow_logs(connection, op.get_context().opts['aiida_profile'])

    # Note, we could also remove objpk and objname from the metadata dictionary here,
    # but since this is not yet a JSONB column, it would be a costly operation, so we skip it for now.

    # Create a new column, which is a foreign key to the dbnode table
    op.add_column(
        'db_dblog', sa.Column('dbnode_id', sa.INTEGER(), autoincrement=False, nullable=False, server_default='1')
    )

    # Transfer data to dbnode_id from objpk
    connection.execute(sa.text("""UPDATE db_dblog SET dbnode_id=objpk"""))

    # Create the foreign key constraint and index
    op.create_foreign_key(
        'db_dblog_dbnode_id_da34b732_fk_db_dbnode_id',
        'db_dblog',
        'db_dbnode',
        ['dbnode_id'],
        ['id'],
        initially='DEFERRED',
        deferrable=True,
        # note, the django migration added on_delete='CASCADE', however, this does not actually set it on the database,
        # see: https://stackoverflow.com/a/35780859/5033292
    )
    op.create_index('db_dblog_dbnode_id_da34b732', 'db_dblog', ['dbnode_id'], unique=False)

    # Now that all the data have been migrated, remove the server default, and unnecessary columns
    op.alter_column('db_dblog', 'dbnode_id', server_default=None)
    op.drop_column('db_dblog', 'objpk')
    op.drop_column('db_dblog', 'objname')

    # Create the UUID column, with a default UUID value
    op.add_column(
        'db_dblog',
        sa.Column(
            'uuid',
            postgresql.UUID(),
            nullable=False,
            server_default='f6a16ff7-4a31-11eb-be7b-8344edc8f36b',
        ),
    )
    op.alter_column('db_dblog', 'uuid', server_default=None)

    # Set unique uuids on the column rows
    set_new_uuid(connection)

    # we now want to set the unique constraint
    # however, this gives: cannot ALTER TABLE "db_dblog" because it has pending trigger events
    # so we do this in a follow up migration (which takes place in a new transaction)
    # op.create_unique_constraint('db_dblog_uuid_9cf77df3_uniq', 'db_dblog', ['uuid'])


def downgrade():
    """Migrations for the downgrade."""
    raise NotImplementedError('Downgrade of django_0024a.')



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Data plugin that models a folder on a remote computer."""

import os

from aiida.orm import AuthInfo
from aiida.orm.fields import add_field

from ..data import Data

__all__ = ('RemoteData',)


class RemoteData(Data):
    """Store a link to a file or folder on a remote machine.

    Remember to pass a computer!
    """

    KEY_EXTRA_CLEANED = 'cleaned'
    __qb_fields__ = [
        add_field(
            'remote_path',
            dtype=str,
        ),
    ]

    def __init__(self, remote_path=None, **kwargs):
        super().__init__(**kwargs)
        if remote_path is not None:
            self.set_remote_path(remote_path)

    def get_remote_path(self) -> str:
        return self.base.attributes.get('remote_path')

    def set_remote_path(self, val):
        self.base.attributes.set('remote_path', val)

    @property
    def is_cleaned(self):
        """Return whether the remote folder has been cleaned."""
        return self.base.extras.get(self.KEY_EXTRA_CLEANED, False)

    @property
    def is_empty(self):
        """Check if remote folder is empty"""
        if self.is_cleaned:
            return True

        authinfo = self.get_authinfo()
        transport = authinfo.get_transport()

        with transport:
            try:
                transport.chdir(self.get_remote_path())
            except OSError:
                # If the transport OSError the directory no longer exists and was deleted
                return True

            return not transport.listdir()

    def getfile(self, relpath, destpath):
        """Connects to the remote folder and retrieves the content of a file.

        :param relpath:  The relative path of the file on the remote to retrieve.
        :param destpath: The absolute path of where to store the file on the local machine.
        """
        authinfo = self.get_authinfo()

        with authinfo.get_transport() as transport:
            try:
                full_path = os.path.join(self.get_remote_path(), relpath)
                transport.getfile(full_path, destpath)
            except OSError as exception:
                if exception.errno == 2:  # file does not exist
                    raise OSError(
                        'The required remote file {} on {} does not exist or has been deleted.'.format(
                            full_path, self.computer.label
                        )
                    ) from exception
                raise

    def listdir(self, relpath='.'):
        """Connects to the remote folder and lists the directory content.

        :param relpath: If 'relpath' is specified, lists the content of the given subfolder.
        :return: a flat list of file/directory names (as strings).
        """
        authinfo = self.get_authinfo()

        with authinfo.get_transport() as transport:
            try:
                full_path = os.path.join(self.get_remote_path(), relpath)
                transport.chdir(full_path)
            except OSError as exception:
                if exception.errno in (2, 20):  # directory not existing or not a directory
                    exc = OSError(
                        f'The required remote folder {full_path} on {self.computer.label} does not exist, is not a '
                        'directory or has been deleted.'
                    )
                    exc.errno = exception.errno
                    raise exc from exception
                else:
                    raise

            try:
                return transport.listdir()
            except OSError as exception:
                if exception.errno in (2, 20):  # directory not existing or not a directory
                    exc = OSError(
                        f'The required remote folder {full_path} on {self.computer.label} does not exist, is not a '
                        'directory or has been deleted.'
                    )
                    exc.errno = exception.errno
                    raise exc from exception
                else:
                    raise

    def listdir_withattributes(self, path='.'):
        """Connects to the remote folder and lists the directory content.

        :param relpath: If 'relpath' is specified, lists the content of the given subfolder.
        :return: a list of dictionaries, where the documentation is in :py:class:Transport.listdir_withattributes.
        """
        authinfo = self.get_authinfo()

        with authinfo.get_transport() as transport:
            try:
                full_path = os.path.join(self.get_remote_path(), path)
                transport.chdir(full_path)
            except OSError as exception:
                if exception.errno in (2, 20):  # directory not existing or not a directory
                    exc = OSError(
                        f'The required remote folder {full_path} on {self.computer.label} does not exist, is not a '
                        'directory or has been deleted.'
                    )
                    exc.errno = exception.errno
                    raise exc from exception
                else:
                    raise

            try:
                return transport.listdir_withattributes()
            except OSError as exception:
                if exception.errno in (2, 20):  # directory not existing or not a directory
                    exc = OSError(
                        f'The required remote folder {full_path} on {self.computer.label} does not exist, is not a '
                        'directory or has been deleted.'
                    )
                    exc.errno = exception.errno
                    raise exc from exception
                else:
                    raise

    def _clean(self, transport=None):
        """Remove all content of the remote folder on the remote computer.

        When the cleaning operation is successful, the extra with the key ``RemoteData.KEY_EXTRA_CLEANED`` is set.

        :param transport: Provide an optional transport that is already open. If not provided, a transport will be
            automatically opened, based on the current default user and the computer of this data node. Passing in the
            transport can be used for efficiency if a great number of nodes need to be cleaned for the same computer.
            Note that the user should take care that the correct transport is passed.
        :raises ValueError: If the hostname of the provided transport does not match that of the node's computer.
        """
        from aiida.orm.utils.remote import clean_remote

        remote_dir = self.get_remote_path()

        if transport is None:
            with self.get_authinfo().get_transport() as transport:
                clean_remote(transport, remote_dir)
        else:
            if transport.hostname != self.computer.hostname:
                raise ValueError(
                    f'Transport hostname `{transport.hostname}` does not equal `{self.computer.hostname}` of {self}.'
                )
            clean_remote(transport, remote_dir)

        self.base.extras.set(self.KEY_EXTRA_CLEANED, True)

    def _validate(self):
        from aiida.common.exceptions import ValidationError

        super()._validate()

        try:
            self.get_remote_path()
        except AttributeError as exception:
            raise ValidationError("attribute 'remote_path' not set.") from exception

        computer = self.computer
        if computer is None:
            raise ValidationError('Remote computer not set.')

    def get_authinfo(self):
        return AuthInfo.get_collection(self.backend).get(dbcomputer=self.computer, aiidauser=self.user)



"""Interface for comments of a node instance."""

from __future__ import annotations

import typing as t

from ..comments import Comment
from ..users import User

if t.TYPE_CHECKING:
    from .node import Node


class NodeComments:
    """Interface for comments of a node instance."""

    def __init__(self, node: 'Node') -> None:
        """Initialize the comments interface."""
        self._node = node

    def add(self, content: str, user: t.Optional[User] = None) -> Comment:
        """Add a new comment.

        :param content: string with comment
        :param user: the user to associate with the comment, will use default if not supplied
        :return: the newly created comment
        """
        user = user or User.get_collection(self._node.backend).get_default()
        return Comment(backend=self._node.backend, node=self._node, user=user, content=content).store()

    def get(self, identifier: int) -> Comment:
        """Return a comment corresponding to the given identifier.

        :param identifier: the comment pk
        :raise aiida.common.NotExistent: if the comment with the given id does not exist
        :raise aiida.common.MultipleObjectsError: if the id cannot be uniquely resolved to a comment
        :return: the comment
        """
        return Comment.get_collection(self._node.backend).get(dbnode_id=self._node.pk, id=identifier)

    def all(self) -> list[Comment]:
        """Return a sorted list of comments for this node.

        :return: the list of comments, sorted by pk
        """
        return Comment.get_collection(self._node.backend).find(
            filters={'dbnode_id': self._node.pk}, order_by=[{'id': 'asc'}]
        )

    def update(self, identifier: int, content: str) -> None:
        """Update the content of an existing comment.

        :param identifier: the comment pk
        :param content: the new comment content
        :raise aiida.common.NotExistent: if the comment with the given id does not exist
        :raise aiida.common.MultipleObjectsError: if the id cannot be uniquely resolved to a comment
        """
        comment = Comment.get_collection(self._node.backend).get(dbnode_id=self._node.pk, id=identifier)
        comment.set_content(content)

    def remove(self, identifier: int) -> None:
        """Delete an existing comment.

        :param identifier: the comment pk
        """
        Comment.get_collection(self._node.backend).delete(identifier)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`verdi code` command."""

from collections import defaultdict
from functools import partial

import click

from aiida.cmdline.commands.cmd_verdi import verdi
from aiida.cmdline.groups.dynamic import DynamicEntryPointCommandGroup
from aiida.cmdline.params import arguments, options, types
from aiida.cmdline.params.options.commands import code as options_code
from aiida.cmdline.utils import echo, echo_tabulate
from aiida.cmdline.utils.decorators import with_dbenv
from aiida.common import exceptions


@verdi.group('code')
def verdi_code():
    """Setup and manage codes."""


def create_code(ctx: click.Context, cls, non_interactive: bool, **kwargs):
    """Create a new `Code` instance."""
    try:
        instance = cls(**kwargs)
    except (TypeError, ValueError) as exception:
        echo.echo_critical(f'Failed to create instance `{cls}`: {exception}')

    try:
        instance.store()
    except exceptions.ValidationError as exception:
        echo.echo_critical(f'Failed to store instance of `{cls}`: {exception}')

    echo.echo_success(f'Created {cls.__name__}<{instance.pk}>')


@verdi_code.group(
    'create',
    cls=DynamicEntryPointCommandGroup,
    command=create_code,
    entry_point_group='aiida.data',
    entry_point_name_filter=r'core\.code\..*',
)
def code_create():
    """Create a new code."""


def get_default(key, ctx):
    """Get the default argument using a user instance property
    :param value: The name of the property to use
    :param ctx: The click context (which will be used to get the user)
    :return: The default value, or None
    """
    try:
        value = getattr(ctx.code_builder, key)
        if value == '':
            value = None
    except KeyError:
        value = None

    return value


def get_computer_name(ctx):
    return getattr(ctx.code_builder, 'computer').label


def get_on_computer(ctx):
    return not getattr(ctx.code_builder, 'is_local')()


def set_code_builder(ctx, param, value):
    """Set the code spec for defaults of following options."""
    from aiida.orm.utils.builders.code import CodeBuilder

    ctx.code_builder = CodeBuilder.from_code(value)
    return value


# Defining the ``COMPUTER`` option first guarantees that the user is prompted for the computer first. This is necessary
# because the ``LABEL`` option has a callback that relies on the computer being already set. Execution order is
# guaranteed only for the interactive case, however. For the non-interactive case, the callback is called explicitly
# once more in the command body to cover the case when the label is specified before the computer.
@verdi_code.command('setup', deprecated='Please use `verdi code create` instead.')
@options_code.ON_COMPUTER()
@options_code.COMPUTER()
@options_code.LABEL()
@options_code.DESCRIPTION()
@options_code.INPUT_PLUGIN()
@options_code.REMOTE_ABS_PATH()
@options_code.FOLDER()
@options_code.REL_PATH()
@options_code.USE_DOUBLE_QUOTES()
@options_code.PREPEND_TEXT()
@options_code.APPEND_TEXT()
@options.NON_INTERACTIVE()
@options.CONFIG_FILE()
@click.pass_context
@with_dbenv()
def setup_code(ctx, non_interactive, **kwargs):
    """Setup a new code (use `verdi code create`)."""
    from aiida.orm.utils.builders.code import CodeBuilder

    options_code.validate_label_uniqueness(ctx, None, kwargs['label'])

    if kwargs.pop('on_computer'):
        kwargs['code_type'] = CodeBuilder.CodeType.ON_COMPUTER
    else:
        kwargs['code_type'] = CodeBuilder.CodeType.STORE_AND_UPLOAD

    # Convert entry point to its name
    if kwargs['input_plugin']:
        kwargs['input_plugin'] = kwargs['input_plugin'].name

    code_builder = CodeBuilder(**kwargs)

    try:
        code = code_builder.new()
    except ValueError as exception:
        echo.echo_critical(f'invalid inputs: {exception}')

    try:
        code.store()
    except Exception as exception:
        echo.echo_critical(f'Unable to store the Code: {exception}')

    echo.echo_success(f'Code<{code.pk}> {code.full_label} created')


@verdi_code.command('test')
@arguments.CODE(callback=set_code_builder)
@with_dbenv()
def code_test(code):
    """Run tests for the given code to check whether it is usable.

    For remote codes the following checks are performed:

     * Whether the remote executable exists.

    """
    from aiida.orm import InstalledCode

    if isinstance(code, InstalledCode):
        try:
            code.validate_filepath_executable()
        except exceptions.ValidationError as exception:
            echo.echo_critical(f'validation failed: {exception}')

    echo.echo_success('all tests succeeded.')


# Defining the ``COMPUTER`` option first guarantees that the user is prompted for the computer first. This is necessary
# because the ``LABEL`` option has a callback that relies on the computer being already set. Execution order is
# guaranteed only for the interactive case, however. For the non-interactive case, the callback is called explicitly
# once more in the command body to cover the case when the label is specified before the computer.
@verdi_code.command('duplicate')
@arguments.CODE(callback=set_code_builder)
@options_code.ON_COMPUTER(contextual_default=get_on_computer)
@options_code.COMPUTER(contextual_default=get_computer_name)
@options_code.LABEL(contextual_default=partial(get_default, 'label'))
@options_code.DESCRIPTION(contextual_default=partial(get_default, 'description'))
@options_code.INPUT_PLUGIN(contextual_default=partial(get_default, 'input_plugin'))
@options_code.REMOTE_ABS_PATH(contextual_default=partial(get_default, 'remote_abs_path'))
@options_code.FOLDER(contextual_default=partial(get_default, 'code_folder'))
@options_code.REL_PATH(contextual_default=partial(get_default, 'code_rel_path'))
@options_code.PREPEND_TEXT(contextual_default=partial(get_default, 'prepend_text'))
@options_code.APPEND_TEXT(contextual_default=partial(get_default, 'append_text'))
@options.NON_INTERACTIVE()
@click.option('--hide-original', is_flag=True, default=False, help='Hide the code being copied.')
@click.pass_context
@with_dbenv()
def code_duplicate(ctx, code, non_interactive, **kwargs):
    """Duplicate a code allowing to change some parameters."""
    from aiida.common.exceptions import ValidationError
    from aiida.orm.utils.builders.code import CodeBuilder

    options_code.validate_label_uniqueness(ctx, None, kwargs['label'])

    if kwargs.pop('on_computer'):
        kwargs['code_type'] = CodeBuilder.CodeType.ON_COMPUTER
    else:
        kwargs['code_type'] = CodeBuilder.CodeType.STORE_AND_UPLOAD

    if kwargs.pop('hide_original'):
        code.is_hidden = True

    # Convert entry point to its name
    if kwargs['input_plugin']:
        kwargs['input_plugin'] = kwargs['input_plugin'].name

    code_builder = ctx.code_builder
    for key, value in kwargs.items():
        setattr(code_builder, key, value)
    new_code = code_builder.new()

    try:
        new_code.store()
        new_code.is_hidden = False
    except ValidationError as exception:
        echo.echo_critical(f'Unable to store the Code: {exception}')

    echo.echo_success(f'Code<{new_code.pk}> {new_code.full_label} created')


@verdi_code.command()
@arguments.CODE()
@with_dbenv()
def show(code):
    """Display detailed information for a code."""
    from aiida.cmdline import is_verbose

    table = []
    table.append(['PK', code.pk])
    table.append(['UUID', code.uuid])
    table.append(['Type', code.entry_point.name])
    for key in code.Model.model_fields.keys():
        try:
            table.append([key.capitalize().replace('_', ' '), getattr(code, key)])
        except AttributeError:
            continue
    if is_verbose():
        table.append(['Calculations', len(code.base.links.get_outgoing().all())])

    echo_tabulate(table)


@verdi_code.command()
@arguments.CODE()
@arguments.OUTPUT_FILE(type=click.Path(exists=False))
@click.option(
    '--sort/--no-sort',
    is_flag=True,
    default=True,
    help='Sort the keys of the output YAML.',
    show_default=True,
)
@with_dbenv()
def export(code, output_file, sort):
    """Export code to a yaml file."""
    import yaml

    code_data = {}

    for key in code.Model.model_fields.keys():
        if key == 'computer':
            value = getattr(code, key).label
        else:
            value = getattr(code, key)

        # If the attribute is not set, for example ``with_mpi`` do not export it, because the YAML won't be valid for
        # use in ``verdi code create`` since ``None`` is not a valid value on the CLI.
        if value is not None:
            code_data[key] = str(value)

    with open(output_file, 'w', encoding='utf-8') as yfhandle:
        yaml.dump(code_data, yfhandle, sort_keys=sort)


@verdi_code.command()
@arguments.CODES()
@options.DRY_RUN()
@options.FORCE()
@with_dbenv()
def delete(codes, dry_run, force):
    """Delete a code.

    Note that codes are part of the data provenance, and deleting a code will delete all calculations using it.
    """
    from aiida.tools import delete_nodes

    node_pks_to_delete = [code.pk for code in codes]

    def _dry_run_callback(pks):
        if not pks or force:
            return False
        echo.echo_warning(f'YOU ARE ABOUT TO DELETE {len(pks)} NODES! THIS CANNOT BE UNDONE!')
        return not click.confirm('Shall I continue?', abort=True)

    was_deleted = delete_nodes(node_pks_to_delete, dry_run=dry_run or _dry_run_callback)

    if was_deleted:
        echo.echo_success('Finished deletion.')


@verdi_code.command()
@arguments.CODES()
@with_dbenv()
def hide(codes):
    """Hide one or more codes from `verdi code list`."""
    for code in codes:
        code.is_hidden = True
        echo.echo_success(f'Code<{code.pk}> {code.full_label} hidden')


@verdi_code.command()
@arguments.CODES()
@with_dbenv()
def reveal(codes):
    """Reveal one or more hidden codes in `verdi code list`."""
    for code in codes:
        code.is_hidden = False
        echo.echo_success(f'Code<{code.pk}> {code.full_label} revealed')


@verdi_code.command()
@arguments.CODE()
@arguments.LABEL()
@with_dbenv()
def relabel(code, label):
    """Relabel a code."""
    old_label = code.full_label

    try:
        code.label = label
    except (TypeError, ValueError) as exception:
        echo.echo_critical(f'invalid code label: {exception}')
    else:
        echo.echo_success(f'Code<{code.pk}> relabeled from {old_label} to {code.full_label}')


VALID_PROJECTIONS = {
    'full_label': [('code', 'label'), ('computer', 'label')],
    'entry_point': [('code', 'node_type')],
    'pk': [('code', 'id')],
    'uuid': [('code', 'uuid')],
    'label': [('code', 'label')],
    'default_calc_job_plugin': [('code', 'attributes.input_plugin')],
    'computer': [('computer', 'label')],
    'user': [('user', 'email')],
}


@verdi_code.command('list')
@options.COMPUTER(help='Filter codes by computer.')
@click.option(
    '-d',
    '--default-calc-job-plugin',
    type=types.PluginParamType(group='calculations', load=False),
    help='Filter codes by their optional default calculation job plugin.',
)
@options.ALL(help='Include hidden codes.')
@options.ALL_USERS(help='Include codes from all users.')
@options.PROJECT(type=click.Choice(VALID_PROJECTIONS.keys()), default=['full_label', 'pk', 'entry_point'])
@options.RAW()
@click.option('-o', '--show-owner', 'show_owner', is_flag=True, default=False, help='Show owners of codes.')
@with_dbenv()
def code_list(computer, default_calc_job_plugin, all_entries, all_users, raw, show_owner, project):
    """List the available codes."""
    from aiida import orm
    from aiida.orm.utils.node import load_node_class

    if show_owner:
        echo.echo_deprecated(
            'the `-o/--show-owner` option is deprecated. To show the user use the `-P/--project` option instead, e.g., '
            '`verdi code list -P full_label user`.'
        )
        if 'user' not in project:
            project = project + ('user',)

    filters = defaultdict(dict)
    projections = defaultdict(list)

    for key in project:
        for entity, projection in VALID_PROJECTIONS[key]:
            if projection not in projections[entity]:
                projections[entity].append(projection)

    if not all_entries:
        filters['code'][f'extras.{orm.Code.HIDDEN_KEY}'] = {'!==': True}

    if not all_users:
        filters['user']['email'] = orm.User.collection.get_default().email

    if computer is not None:
        filters['computer']['uuid'] = computer.uuid

    if default_calc_job_plugin is not None:
        filters['code']['attributes.input_plugin'] = default_calc_job_plugin.name

    query = orm.QueryBuilder()
    query.append(orm.Code, tag='code', project=projections.get('code', None), filters=filters.get('code', None))
    query.append(
        orm.Computer,
        tag='computer',
        with_node='code',
        project=projections.get('computer', None),
        filters=filters.get('computer', None),
    )
    query.append(
        orm.User, tag='user', with_node='code', project=projections.get('user', None), filters=filters.get('user', None)
    )
    query.order_by({'code': {'id': 'asc'}})

    if not query.count():
        echo.echo_report('No codes found matching the specified criteria.')
        return

    table = []
    headers = [projection.replace('_', ' ').capitalize() for projection in project] if not raw else []
    table_format = 'plain' if raw else None

    for result in query.iterdict():
        row = []
        for key in project:
            if key == 'entry_point':
                node_type = result['code']['node_type']
                entry_point = load_node_class(node_type).entry_point
                row.append(entry_point.name)
            else:
                row.append('@'.join(str(result[entity][projection]) for entity, projection in VALID_PROJECTIONS[key]))
        table.append(row)

    echo_tabulate(table, headers=headers, tablefmt=table_format)

    if not raw:
        echo.echo_report('\nUse `verdi code show IDENTIFIER` to see details for a code', prefix=False)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# ruff: noqa: N802
"""A module containing the logic for creating joined queries."""

from dataclasses import dataclass
from typing import Any, Callable, Dict, Optional, Protocol, Type

from sqlalchemy import and_, join, select
from sqlalchemy.dialects.postgresql import array
from sqlalchemy.orm import Query, aliased
from sqlalchemy.orm.util import AliasedClass
from sqlalchemy.sql.elements import BooleanClauseList
from sqlalchemy.sql.expression import cast as type_cast
from sqlalchemy.sql.schema import Table
from sqlalchemy.types import Integer

from aiida.common.links import LinkType
from aiida.storage.psql_dos.models.base import Model


class _EntityMapper(Protocol):
    """Mapping of implemented entity types."""

    @property
    def AuthInfo(self) -> Type[Model]: ...

    @property
    def Node(self) -> Type[Model]: ...

    @property
    def Group(self) -> Type[Model]: ...

    @property
    def Link(self) -> Type[Model]: ...

    @property
    def User(self) -> Type[Model]: ...

    @property
    def Computer(self) -> Type[Model]: ...

    @property
    def Comment(self) -> Type[Model]: ...

    @property
    def Log(self) -> Type[Model]: ...

    @property
    def table_groups_nodes(self) -> Type[Table]: ...


@dataclass
class JoinReturn:
    join: Callable[[Query], Query]
    aliased_edge: Optional[AliasedClass] = None
    edge_tag: str = ''


FilterType = Dict[str, Any]
JoinFuncType = Callable[[Any, Any, bool, FilterType, bool], JoinReturn]


class SqlaJoiner:
    """A class containing the logic for SQLAlchemy entities joining entities."""

    def __init__(
        self,
        entity_mapper: _EntityMapper,
        filter_builder: Callable[[AliasedClass, FilterType], Optional[BooleanClauseList]],
    ):
        """Initialise the class"""
        self._entities = entity_mapper
        self._build_filters = filter_builder

    def get_join_func(self, entity_key: str, relationship: str) -> JoinFuncType:
        """Return the function to join two entities"""
        return self._entity_join_map()[entity_key][relationship]

    def _entity_join_map(self) -> Dict[str, Dict[str, JoinFuncType]]:
        """Map relationship type keywords to functions
        The first level defines the entity which has been passed to the qb.append function,
        and the second defines the relationship with respect to a given tag.
        """
        mapping = {
            'authinfo': {
                'with_computer': self._join_computer_authinfo,
                'with_user': self._join_user_authinfo,
            },
            'comment': {
                'with_node': self._join_node_comment,
                'with_user': self._join_user_comment,
            },
            'computer': {
                'with_node': self._join_node_computer,
            },
            'group': {
                'with_node': self._join_node_group,
                'with_user': self._join_user_group,
            },
            'link': {},
            'log': {
                'with_node': self._join_node_log,
            },
            'node': {
                'with_log': self._join_log_node,
                'with_comment': self._join_comment_node,
                'with_incoming': self._join_node_outputs,
                'with_outgoing': self._join_node_inputs,
                'with_descendants': self._join_node_ancestors_recursive,
                'with_ancestors': self._join_node_descendants_recursive,
                'with_computer': self._join_computer_node,
                'with_user': self._join_user_node,
                'with_group': self._join_group_node,
            },
            'user': {
                'with_authinfo': self._join_authinfo_user,
                'with_comment': self._join_comment_user,
                'with_node': self._join_node_user,
                'with_group': self._join_group_user,
            },
        }

        return mapping  # type: ignore

    def _join_computer_authinfo(self, joined_entity, entity_to_join, isouterjoin: bool, **_kw):
        """:param joined_entity: the aliased user you want to join to
        :param entity_to_join: the (aliased) node or group in the DB to join with
        """
        _check_dbentities(
            (joined_entity, self._entities.Computer), (entity_to_join, self._entities.AuthInfo), 'with_computer'
        )

        def new_query(q):
            return q.join(entity_to_join, entity_to_join.dbcomputer_id == joined_entity.id, isouter=isouterjoin)

        return JoinReturn(new_query)

    def _join_user_authinfo(self, joined_entity, entity_to_join, isouterjoin: bool, **_kw):
        """:param joined_entity: the aliased user you want to join to
        :param entity_to_join: the (aliased) node or group in the DB to join with
        """
        _check_dbentities((joined_entity, self._entities.User), (entity_to_join, self._entities.AuthInfo), 'with_user')

        def new_query(q):
            return q.join(entity_to_join, entity_to_join.aiidauser_id == joined_entity.id, isouter=isouterjoin)

        return JoinReturn(new_query)

    def _join_group_node(self, joined_entity, entity_to_join, isouterjoin: bool, **_kw):
        """:param joined_entity:
            The (aliased) ORMclass that is
            a group in the database
        :param entity_to_join:
            The (aliased) ORMClass that is a node and member of the group

        **joined_entity** and **entity_to_join**
        are joined via the table_groups_nodes table.
        from **joined_entity** as group to **enitity_to_join** as node.
        (**enitity_to_join** is *with_group* **joined_entity**)
        """
        _check_dbentities((joined_entity, self._entities.Group), (entity_to_join, self._entities.Node), 'with_group')
        aliased_group_nodes = aliased(self._entities.table_groups_nodes)

        def new_query(q):
            return q.join(aliased_group_nodes, aliased_group_nodes.c.dbgroup_id == joined_entity.id).join(
                entity_to_join, entity_to_join.id == aliased_group_nodes.c.dbnode_id, isouter=isouterjoin
            )

        return JoinReturn(new_query, aliased_group_nodes)

    def _join_node_group(self, joined_entity, entity_to_join, isouterjoin: bool, **_kw):
        """:param joined_entity: The (aliased) node in the database
        :param entity_to_join: The (aliased) Group

        **joined_entity** and **entity_to_join** are
        joined via the table_groups_nodes table.
        from **joined_entity** as node to **enitity_to_join** as group.
        (**enitity_to_join** is a group *with_node* **joined_entity**)
        """
        _check_dbentities((joined_entity, self._entities.Node), (entity_to_join, self._entities.Group), 'with_node')
        aliased_group_nodes = aliased(self._entities.table_groups_nodes)

        def new_query(q):
            return q.join(aliased_group_nodes, aliased_group_nodes.c.dbnode_id == joined_entity.id).join(
                entity_to_join, entity_to_join.id == aliased_group_nodes.c.dbgroup_id, isouter=isouterjoin
            )

        return JoinReturn(new_query, aliased_group_nodes)

    def _join_node_user(self, joined_entity, entity_to_join, isouterjoin: bool, **_kw):
        """:param joined_entity: the aliased node
        :param entity_to_join: the aliased user to join to that node
        """
        _check_dbentities((joined_entity, self._entities.Node), (entity_to_join, self._entities.User), 'with_node')

        def new_query(q):
            return q.join(entity_to_join, entity_to_join.id == joined_entity.user_id, isouter=isouterjoin)

        return JoinReturn(new_query)

    def _join_user_node(self, joined_entity, entity_to_join, isouterjoin: bool, **_kw):
        """:param joined_entity: the aliased user you want to join to
        :param entity_to_join: the (aliased) node or group in the DB to join with
        """
        _check_dbentities((joined_entity, self._entities.User), (entity_to_join, self._entities.Node), 'with_user')

        def new_query(q):
            return q.join(entity_to_join, entity_to_join.user_id == joined_entity.id, isouter=isouterjoin)

        return JoinReturn(new_query)

    def _join_computer_node(self, joined_entity, entity_to_join, isouterjoin: bool, **_kw):
        """:param joined_entity: the (aliased) computer entity
        :param entity_to_join: the (aliased) node entity

        """
        _check_dbentities(
            (joined_entity, self._entities.Computer), (entity_to_join, self._entities.Node), 'with_computer'
        )

        def new_query(q):
            return q.join(entity_to_join, entity_to_join.dbcomputer_id == joined_entity.id, isouter=isouterjoin)

        return JoinReturn(new_query)

    def _join_node_computer(self, joined_entity, entity_to_join, isouterjoin: bool, **_kw):
        """:param joined_entity: An entity that can use a computer (eg a node)
        :param entity_to_join: aliased dbcomputer entity
        """
        _check_dbentities((joined_entity, self._entities.Node), (entity_to_join, self._entities.Computer), 'with_node')

        def new_query(q):
            return q.join(entity_to_join, joined_entity.dbcomputer_id == entity_to_join.id, isouter=isouterjoin)

        return JoinReturn(new_query)

    def _join_group_user(self, joined_entity, entity_to_join, isouterjoin: bool, **_kw):
        """:param joined_entity: An aliased dbgroup
        :param entity_to_join: aliased dbuser
        """
        _check_dbentities((joined_entity, self._entities.Group), (entity_to_join, self._entities.User), 'with_group')

        def new_query(q):
            return q.join(entity_to_join, joined_entity.user_id == entity_to_join.id, isouter=isouterjoin)

        return JoinReturn(new_query)

    def _join_user_group(self, joined_entity, entity_to_join, isouterjoin: bool, **_kw):
        """:param joined_entity: An aliased user
        :param entity_to_join: aliased group
        """
        _check_dbentities((joined_entity, self._entities.User), (entity_to_join, self._entities.Group), 'with_user')

        def new_query(q):
            return q.join(entity_to_join, joined_entity.id == entity_to_join.user_id, isouter=isouterjoin)

        return JoinReturn(new_query)

    def _join_node_comment(self, joined_entity, entity_to_join, isouterjoin: bool, **_kw):
        """:param joined_entity: An aliased node
        :param entity_to_join: aliased comment
        """
        _check_dbentities((joined_entity, self._entities.Node), (entity_to_join, self._entities.Comment), 'with_node')

        def new_query(q):
            return q.join(entity_to_join, joined_entity.id == entity_to_join.dbnode_id, isouter=isouterjoin)

        return JoinReturn(new_query)

    def _join_comment_node(self, joined_entity, entity_to_join, isouterjoin: bool, **_kw):
        """:param joined_entity: An aliased comment
        :param entity_to_join: aliased node
        """
        _check_dbentities(
            (joined_entity, self._entities.Comment), (entity_to_join, self._entities.Node), 'with_comment'
        )

        def new_query(q):
            return q.join(entity_to_join, joined_entity.dbnode_id == entity_to_join.id, isouter=isouterjoin)

        return JoinReturn(new_query)

    def _join_node_log(self, joined_entity, entity_to_join, isouterjoin: bool, **_kw):
        """:param joined_entity: An aliased node
        :param entity_to_join: aliased log
        """
        _check_dbentities((joined_entity, self._entities.Node), (entity_to_join, self._entities.Log), 'with_node')

        def new_query(q):
            return q.join(entity_to_join, joined_entity.id == entity_to_join.dbnode_id, isouter=isouterjoin)

        return JoinReturn(new_query)

    def _join_log_node(self, joined_entity, entity_to_join, isouterjoin: bool, **_kw):
        """:param joined_entity: An aliased log
        :param entity_to_join: aliased node
        """
        _check_dbentities((joined_entity, self._entities.Log), (entity_to_join, self._entities.Node), 'with_log')

        def new_query(q):
            return q.join(entity_to_join, joined_entity.dbnode_id == entity_to_join.id, isouter=isouterjoin)

        return JoinReturn(new_query)

    def _join_user_comment(self, joined_entity, entity_to_join, isouterjoin: bool, **_kw):
        """:param joined_entity: An aliased user
        :param entity_to_join: aliased comment
        """
        _check_dbentities((joined_entity, self._entities.User), (entity_to_join, self._entities.Comment), 'with_user')

        def new_query(q):
            return q.join(entity_to_join, joined_entity.id == entity_to_join.user_id, isouter=isouterjoin)

        return JoinReturn(new_query)

    def _join_authinfo_user(self, joined_entity, entity_to_join, isouterjoin: bool, **_kw):
        """:param joined_entity: An aliased comment
        :param entity_to_join: aliased user
        """
        _check_dbentities(
            (joined_entity, self._entities.AuthInfo), (entity_to_join, self._entities.User), 'with_authinfo'
        )

        def new_query(q):
            return q.join(entity_to_join, joined_entity.aiidauser_id == entity_to_join.id, isouter=isouterjoin)

        return JoinReturn(new_query)

    def _join_comment_user(self, joined_entity, entity_to_join, isouterjoin: bool, **_kw):
        """:param joined_entity: An aliased comment
        :param entity_to_join: aliased user
        """
        _check_dbentities(
            (joined_entity, self._entities.Comment), (entity_to_join, self._entities.User), 'with_comment'
        )

        def new_query(q):
            return q.join(entity_to_join, joined_entity.user_id == entity_to_join.id, isouter=isouterjoin)

        return JoinReturn(new_query)

    def _join_node_outputs(self, joined_entity, entity_to_join, isouterjoin: bool, **_kw):
        """:param joined_entity: The (aliased) ORMclass that is an input
        :param entity_to_join: The (aliased) ORMClass that is an output.

        **joined_entity** and **entity_to_join** are joined with a link
        from **joined_entity** as input to **enitity_to_join** as output
        (**enitity_to_join** is *with_incoming* **joined_entity**)
        """
        _check_dbentities((joined_entity, self._entities.Node), (entity_to_join, self._entities.Node), 'with_incoming')

        aliased_edge = aliased(self._entities.Link)

        def new_query(q):
            return q.join(aliased_edge, aliased_edge.input_id == joined_entity.id, isouter=isouterjoin).join(
                entity_to_join, aliased_edge.output_id == entity_to_join.id, isouter=isouterjoin
            )

        return JoinReturn(new_query, aliased_edge)

    def _join_node_inputs(self, joined_entity, entity_to_join, isouterjoin: bool, **_kw):
        """:param joined_entity: The (aliased) ORMclass that is an output
        :param entity_to_join: The (aliased) ORMClass that is an input.

        **joined_entity** and **entity_to_join** are joined with a link
        from **joined_entity** as output to **enitity_to_join** as input
        (**enitity_to_join** is *with_outgoing* **joined_entity**)

        """
        _check_dbentities((joined_entity, self._entities.Node), (entity_to_join, self._entities.Node), 'with_outgoing')
        aliased_edge = aliased(self._entities.Link)

        def new_query(q):
            return q.join(aliased_edge, aliased_edge.output_id == joined_entity.id).join(
                entity_to_join, aliased_edge.input_id == entity_to_join.id, isouter=isouterjoin
            )

        return JoinReturn(new_query, aliased_edge)

    def _join_node_descendants_recursive(
        self, joined_entity, entity_to_join, isouterjoin: bool, filter_dict: FilterType, expand_path=False
    ):
        """Joining descendants using the recursive functionality
        :TODO: Move the filters to be done inside the recursive query (for example on depth)
        :TODO: Pass an option to also show the path, if this is wanted.
        """
        _check_dbentities((joined_entity, self._entities.Node), (entity_to_join, self._entities.Node), 'with_ancestors')

        link1 = aliased(self._entities.Link)
        link2 = aliased(self._entities.Link)
        node1 = aliased(self._entities.Node)

        link_filters = link1.type.in_((LinkType.CREATE.value, LinkType.INPUT_CALC.value))  # follow input / create links
        in_recursive_filters = self._build_filters(node1, filter_dict)
        if in_recursive_filters is None:
            filters = link_filters
        else:
            filters = and_(in_recursive_filters, link_filters)

        selection_walk_list = [
            link1.input_id.label('ancestor_id'),
            link1.output_id.label('descendant_id'),
            type_cast(0, Integer).label('depth'),  # type: ignore[type-var]
        ]
        if expand_path:
            selection_walk_list.append(array((link1.input_id, link1.output_id)).label('path'))

        walk = (
            select(*selection_walk_list)
            .select_from(join(node1, link1, link1.input_id == node1.id))
            .where(filters)
            .cte(recursive=True)
        )

        aliased_walk = aliased(walk)

        selection_union_list = [
            aliased_walk.c.ancestor_id.label('ancestor_id'),
            link2.output_id.label('descendant_id'),
            (aliased_walk.c.depth + type_cast(1, Integer)).label('current_depth'),  # type: ignore[type-var]
        ]
        if expand_path:
            selection_union_list.append((aliased_walk.c.path + array((link2.output_id,))).label('path'))

        descendants_recursive = aliased(
            aliased_walk.union_all(
                select(*selection_union_list)
                .select_from(
                    join(
                        aliased_walk,
                        link2,
                        link2.input_id == aliased_walk.c.descendant_id,
                    )
                )
                .where(link2.type.in_((LinkType.CREATE.value, LinkType.INPUT_CALC.value)))
            )
        )  # .alias()

        def new_query(q):
            return q.join(descendants_recursive, descendants_recursive.c.ancestor_id == joined_entity.id).join(
                entity_to_join, descendants_recursive.c.descendant_id == entity_to_join.id, isouter=isouterjoin
            )

        return JoinReturn(new_query, descendants_recursive.c)

    def _join_node_ancestors_recursive(
        self, joined_entity, entity_to_join, isouterjoin: bool, filter_dict: FilterType, expand_path=False
    ):
        """Joining ancestors using the recursive functionality
        :TODO: Move the filters to be done inside the recursive query (for example on depth)
        :TODO: Pass an option to also show the path, if this is wanted.

        """
        _check_dbentities((joined_entity, self._entities.Node), (entity_to_join, self._entities.Node), 'with_ancestors')

        link1 = aliased(self._entities.Link)
        link2 = aliased(self._entities.Link)
        node1 = aliased(self._entities.Node)

        link_filters = link1.type.in_((LinkType.CREATE.value, LinkType.INPUT_CALC.value))  # follow input / create links
        in_recursive_filters = self._build_filters(node1, filter_dict)
        if in_recursive_filters is None:
            filters = link_filters
        else:
            filters = and_(in_recursive_filters, link_filters)

        selection_walk_list = [
            link1.input_id.label('ancestor_id'),
            link1.output_id.label('descendant_id'),
            type_cast(0, Integer).label('depth'),  # type: ignore[type-var]
        ]
        if expand_path:
            selection_walk_list.append(array((link1.output_id, link1.input_id)).label('path'))

        walk = (
            select(*selection_walk_list)
            .select_from(join(node1, link1, link1.output_id == node1.id))
            .where(filters)
            .cte(recursive=True)
        )

        aliased_walk = aliased(walk)

        selection_union_list = [
            link2.input_id.label('ancestor_id'),
            aliased_walk.c.descendant_id.label('descendant_id'),
            (aliased_walk.c.depth + type_cast(1, Integer)).label('current_depth'),  # type: ignore[type-var]
        ]
        if expand_path:
            selection_union_list.append((aliased_walk.c.path + array((link2.input_id,))).label('path'))

        ancestors_recursive = aliased(
            aliased_walk.union_all(
                select(*selection_union_list)
                .select_from(
                    join(
                        aliased_walk,
                        link2,
                        link2.output_id == aliased_walk.c.ancestor_id,
                    )
                )
                .where(link2.type.in_((LinkType.CREATE.value, LinkType.INPUT_CALC.value)))
                # I can't follow RETURN or CALL links
            )
        )

        def new_query(q):
            return q.join(ancestors_recursive, ancestors_recursive.c.descendant_id == joined_entity.id).join(
                entity_to_join, ancestors_recursive.c.ancestor_id == entity_to_join.id, isouter=isouterjoin
            )

        return JoinReturn(new_query, ancestors_recursive.c)


def _check_dbentities(entities_cls_joined, entities_cls_to_join, relationship: str):
    """Type check for entities

    :param entities_cls_joined:
        A tuple of the aliased class passed as joined_entity and the ormclass that was expected
    :type entities_cls_to_join: tuple
    :param entities_cls_joined:
        A tuple of the aliased class passed as entity_to_join and the ormclass that was expected
    :type entities_cls_to_join: tuple
    :param str relationship:
        The relationship between the two entities to make the Exception comprehensible
    """
    for entity, cls in (entities_cls_joined, entities_cls_to_join):
        if not issubclass(entity._sa_class_manager.class_, cls):
            raise TypeError(
                "You are attempting to join {} as '{}' of {}\n"
                'This failed because you passed:\n'
                ' - {} as entity joined (expected {})\n'
                ' - {} as entity to join (expected {})\n'
                '\n'.format(
                    entities_cls_joined[0].__name__,
                    relationship,
                    entities_cls_to_join[0].__name__,
                    entities_cls_joined[0]._sa_class_manager.class_.__name__,
                    entities_cls_joined[1].__name__,
                    entities_cls_to_join[0]._sa_class_manager.class_.__name__,
                    entities_cls_to_join[1].__name__,
                )
            )



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Local transport"""

###
### GP: a note on the local transport:
### I believe that we must not use os.chdir to keep track of the folder
### in which we are, since this may have very nasty side effects in other
### parts of code, and make things not thread-safe.
### we should instead keep track internally of the 'current working directory'
### in the exact same way as paramiko does already.

import contextlib
import errno
import glob
import io
import os
import shutil
import subprocess

from aiida.transports import cli as transport_cli
from aiida.transports.transport import Transport, TransportInternalError


# refactor or raise the limit: issue #1784
class LocalTransport(Transport):
    """Support copy and command execution on the same host on which AiiDA is running via direct file copy and
    execution commands.

    Note that the environment variables are copied from the submitting process, so you might need to clean it
    with a ``prepend_text``. For example, the AiiDA daemon sets a ``PYTHONPATH``, so you might want to add
    ``unset PYTHONPATH`` if you plan on running calculations that use Python.
    """

    _valid_auth_options = []

    # There is no real limit on how fast you can safely connect to a localhost, unlike often the case with SSH transport
    # where the remote computer will rate limit the number of connections.
    _DEFAULT_SAFE_OPEN_INTERVAL = 0.0

    # Like the connection open interval, for local transport, which should be used for localhost, one probably doesn't
    # want to have a long polling time in most cases. Since localhost is typically used for short running jobs and tests
    # one doesn't want to be waiting tens of seconds for a job that should last less than one second. We don't set 0 but
    # a small finite number that should prevent the CPUs from spinning while still guaranteeing fast throughput.
    DEFAULT_MINIMUM_JOB_POLL_INTERVAL = 0.1

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # The `_internal_dir` will emulate the concept of working directory, as the real current working directory is
        # not to be changed to prevent bug-prone situations
        self._internal_dir = None

        # Just to avoid errors
        self._machine = kwargs.pop('machine', None)
        if self._machine and self._machine != 'localhost':
            self.logger.debug('machine was passed, but it is not localhost')

    def open(self):
        """Opens a local transport channel

        :raise aiida.common.InvalidOperation: if the channel is already open
        """
        from aiida.common.exceptions import InvalidOperation

        if self._is_open:
            raise InvalidOperation('Cannot open the transport twice')

        self._internal_dir = os.path.expanduser('~')
        self._is_open = True
        return self

    def close(self):
        """Closes the local transport channel

        :raise aiida.common.InvalidOperation: if the channel is already open
        """
        from aiida.common.exceptions import InvalidOperation

        if not self._is_open:
            raise InvalidOperation('Cannot close the transport: it is already closed')
        self._is_open = False

    def __str__(self):
        """Return a description as a string."""
        return f"local [{'OPEN' if self._is_open else 'CLOSED'}]"

    @property
    def curdir(self):
        """Returns the _internal_dir, if the channel is open.
        If possible, use getcwd() instead!
        """
        if self._is_open:
            return os.path.realpath(self._internal_dir)

        raise TransportInternalError('Error, local method called for LocalTransport without opening the channel first')

    def chdir(self, path):
        """Changes directory to path, emulated internally.
        :param path: path to cd into
        :raise OSError: if the directory does not have read attributes.
        """
        new_path = os.path.join(self.curdir, path)
        if not os.path.isdir(new_path):
            raise OSError(f"'{new_path}' is not a valid directory")
        elif not os.access(new_path, os.R_OK):
            raise OSError(f"Do not have read permission to '{new_path}'")

        self._internal_dir = os.path.normpath(new_path)

    def chown(self, path, uid, gid):
        os.chown(path, uid, gid)

    def normalize(self, path='.'):
        """Normalizes path, eliminating double slashes, etc..
        :param path: path to normalize
        """
        return os.path.realpath(os.path.join(self.curdir, path))

    def getcwd(self):
        """Returns the current working directory, emulated by the transport"""
        return self.curdir

    @staticmethod
    def _os_path_split_asunder(path):
        """Used by makedirs, Takes path (a str) and returns a list deconcatenating the path."""
        parts = []
        while True:
            newpath, tail = os.path.split(path)
            if newpath == path:
                assert not tail
                if path:
                    parts.append(path)
                break
            parts.append(tail)
            path = newpath
        parts.reverse()
        return parts

    def makedirs(self, path, ignore_existing=False):
        """Super-mkdir; create a leaf directory and all intermediate ones.
        Works like mkdir, except that any intermediate path segment (not
        just the rightmost) will be created if it does not exist.

        :param path: directory to create
        :param ignore_existing: if set to true, it doesn't give any error
                     if the leaf directory does already exist

        :raise OSError: If the directory already exists and is not ignore_existing
        """
        # check to avoid creation of empty dirs
        path = os.path.normpath(path)

        the_path = os.path.join(self.curdir, path)
        to_create = self._os_path_split_asunder(the_path)
        this_dir = ''
        for count, element in enumerate(to_create):
            this_dir = os.path.join(this_dir, element)
            if count + 1 == len(to_create) and self.isdir(this_dir) and ignore_existing:
                return
            if count + 1 == len(to_create) and self.isdir(this_dir) and not ignore_existing:
                os.mkdir(this_dir)
            if not os.path.exists(this_dir):
                os.mkdir(this_dir)

    def mkdir(self, path, ignore_existing=False):
        """Create a folder (directory) named path.

        :param path: name of the folder to create
        :param ignore_existing: if True, does not give any error if the
                directory already exists

        :raise OSError: If the directory already exists.
        """
        if ignore_existing and self.isdir(path):
            return

        os.mkdir(os.path.join(self.curdir, path))

    def rmdir(self, path):
        """Removes a folder at location path.
        :param path: path to remove
        """
        os.rmdir(os.path.join(self.curdir, path))

    def isdir(self, path):
        """Checks if 'path' is a directory.
        :return: a boolean
        """
        if not path:
            return False

        return os.path.isdir(os.path.join(self.curdir, path))

    def chmod(self, path, mode):
        """Changes permission bits of object at path
        :param path: path to modify
        :param mode: permission bits

        :raise OSError: if path does not exist.
        """
        if not path:
            raise OSError('Directory not given in input')
        real_path = os.path.join(self.curdir, path)
        if not os.path.exists(real_path):
            raise OSError('Directory not given in input')
        else:
            os.chmod(real_path, mode)

    # please refactor: issue #1782

    def put(self, localpath, remotepath, *args, **kwargs):
        """Copies a file or a folder from localpath to remotepath.
        Automatically redirects to putfile or puttree.

        :param localpath: absolute path to local file
        :param remotepath: path to remote file
        :param dereference: if True follows symbolic links.
                                 Default = True
        :param overwrite: if True overwrites remotepath.
                                 Default = False

        :raise OSError: if remotepath is not valid
        :raise ValueError: if localpath is not valid
        """
        dereference = kwargs.get('dereference', args[0] if args else True)
        overwrite = kwargs.get('overwrite', args[1] if len(args) > 1 else True)
        ignore_nonexisting = kwargs.get('ignore_noexisting', args[2] if len(args) > 2 else False)
        if not remotepath:
            raise OSError('Input remotepath to put function must be a non empty string')
        if not localpath:
            raise ValueError('Input localpath to put function must be a non empty string')

        if not os.path.isabs(localpath):
            raise ValueError('Source must be an absolute path')

        if self.has_magic(localpath):
            if self.has_magic(remotepath):
                raise ValueError('Pathname patterns are not allowed in the remotepath')

            to_copy_list = glob.glob(localpath)  # using local glob here

            rename_remote = False
            if len(to_copy_list) > 1:
                # I can't scp more than one file on a single file
                if self.isfile(remotepath):
                    raise OSError('Remote remotepath is not a directory')
                # I can't scp more than one file in a non existing directory
                elif not self.path_exists(remotepath):  # questo dovrebbe valere solo per file
                    raise OSError('Remote directory does not exist')
                else:  # the remote path is a directory
                    rename_remote = True

            for source_path in to_copy_list:
                if os.path.isfile(source_path):
                    if rename_remote:  # copying more than one file in one directory
                        # here is the case isfile and more than one file
                        subpath = os.path.join(remotepath, os.path.split(source_path)[1])
                        self.putfile(source_path, subpath, overwrite)

                    elif self.isdir(remotepath):  # one file to copy in '.'
                        subpath = os.path.join(remotepath, os.path.split(source_path)[1])
                        self.putfile(source_path, subpath, overwrite)
                    else:  # one file to copy on one file
                        self.putfile(source_path, remotepath, overwrite)
                else:
                    self.puttree(source_path, remotepath, dereference, overwrite)

        elif os.path.isdir(localpath):
            self.puttree(localpath, remotepath, dereference, overwrite)
        elif os.path.isfile(localpath):
            if self.isdir(remotepath):
                full_destination = os.path.join(remotepath, os.path.split(localpath)[1])
            else:
                full_destination = remotepath

            self.putfile(localpath, full_destination, overwrite)
        elif ignore_nonexisting:
            pass
        else:
            raise OSError(f'The local path {localpath} does not exist')

    def putfile(self, localpath, remotepath, *args, **kwargs):
        """Copies a file from localpath to remotepath.
        Automatically redirects to putfile or puttree.

        :param localpath: absolute path to local file
        :param remotepath: path to remote file
        :param overwrite: if True overwrites remotepath
                                 Default = False

        :raise OSError: if remotepath is not valid
        :raise ValueError: if localpath is not valid
        :raise OSError: if localpath does not exist
        """
        overwrite = kwargs.get('overwrite', args[0] if args else True)
        if not remotepath:
            raise OSError('Input remotepath to putfile must be a non empty string')
        if not localpath:
            raise ValueError('Input localpath to putfile must be a non empty string')

        if not os.path.isabs(localpath):
            raise ValueError('Source must be an absolute path')

        if not os.path.exists(localpath):
            raise OSError('Source does not exists')

        the_destination = os.path.join(self.curdir, remotepath)
        if os.path.exists(the_destination) and not overwrite:
            raise OSError('Destination already exists: not overwriting it')

        shutil.copyfile(localpath, the_destination)

    def puttree(self, localpath, remotepath, *args, **kwargs):
        """Copies a folder recursively from localpath to remotepath.
        Automatically redirects to putfile or puttree.

        :param localpath: absolute path to local file
        :param remotepath: path to remote file
        :param dereference: follow symbolic links.
                                 Default = True
        :param overwrite: if True overwrites remotepath.
                               Default = False

        :raise OSError: if remotepath is not valid
        :raise ValueError: if localpath is not valid
        :raise OSError: if localpath does not exist
        """
        dereference = kwargs.get('dereference', args[0] if args else True)
        overwrite = kwargs.get('overwrite', args[1] if len(args) > 1 else True)
        if not remotepath:
            raise OSError('Input remotepath to putfile must be a non empty string')
        if not localpath:
            raise ValueError('Input localpath to putfile must be a non empty string')

        if not os.path.isabs(localpath):
            raise ValueError('Source must be an absolute path')

        if not os.path.exists(localpath):
            raise OSError('Source does not exists')

        if self.path_exists(remotepath) and not overwrite:
            raise OSError("Can't overwrite existing files")
        if self.isfile(remotepath):
            raise OSError('Cannot copy a directory into a file')

        if self.isdir(remotepath):
            remotepath = os.path.join(remotepath, os.path.split(localpath)[1])

        the_destination = os.path.join(self.curdir, remotepath)

        shutil.copytree(localpath, the_destination, symlinks=not dereference, dirs_exist_ok=overwrite)

    def rmtree(self, path):
        """Remove tree as rm -r would do

        :param path: a string to path
        """
        the_path = os.path.join(self.curdir, path)
        try:
            shutil.rmtree(the_path)
        except OSError as exception:
            if exception.errno == errno.ENOENT:
                pass
            elif exception.errno == errno.ENOTDIR:
                os.remove(the_path)
            else:
                raise OSError(exception)

    # please refactor: issue #1781

    def get(self, remotepath, localpath, *args, **kwargs):
        """Copies a folder or a file recursively from 'remote' remotepath to
        'local' localpath.
        Automatically redirects to getfile or gettree.

        :param remotepath: path to local file
        :param localpath: absolute path to remote file
        :param dereference: follow symbolic links
                                 default = True
        :param overwrite: if True overwrites localpath
                               default = False

        :raise OSError: if 'remote' remotepath is not valid
        :raise ValueError: if 'local' localpath is not valid
        """
        dereference = kwargs.get('dereference', args[0] if args else True)
        overwrite = kwargs.get('overwrite', args[1] if len(args) > 1 else True)
        ignore_nonexisting = kwargs.get('ignore_nonexisting', args[2] if len(args) > 2 else False)
        if not localpath:
            raise ValueError('Input localpath to get function must be a non empty string')
        if not remotepath:
            raise OSError('Input remotepath to get function must be a non empty string')

        if not os.path.isabs(localpath):
            raise ValueError('Destination must be an absolute path')

        if self.has_magic(remotepath):
            if self.has_magic(localpath):
                raise ValueError('Pathname patterns are not allowed in the localpath')
            to_copy_list = self.glob(remotepath)

            rename_local = False
            if len(to_copy_list) > 1:
                # I can't scp more than one file on a single file
                if os.path.isfile(localpath):
                    raise OSError('Remote localpath is not a directory')
                # I can't scp more than one file in a non existing directory
                elif not os.path.exists(localpath):  # this should hold only for files
                    raise OSError('Remote directory does not exist')
                else:  # the remote path is a directory
                    rename_local = True

            for source in to_copy_list:
                if self.isfile(source):
                    if rename_local:  # copying more than one file in one directory
                        # here is the case isfile and more than one file
                        subpath = os.path.join(localpath, os.path.split(source)[1])
                        self.getfile(source, subpath, overwrite)
                    else:  # one file to copy on one file
                        self.getfile(source, localpath, overwrite)
                else:
                    self.gettree(source, localpath, dereference, overwrite)

        elif self.isdir(remotepath):
            self.gettree(remotepath, localpath, dereference, overwrite)
        elif self.isfile(remotepath):
            if os.path.isdir(localpath):
                subpath = os.path.join(localpath, os.path.split(remotepath)[1])
                self.getfile(remotepath, subpath, overwrite)
            else:
                self.getfile(remotepath, localpath, overwrite)
        elif ignore_nonexisting:
            pass
        else:
            raise OSError(f'The remote path {remotepath} does not exist')

    def getfile(self, remotepath, localpath, *args, **kwargs):
        """Copies a file recursively from 'remote' remotepath to
        'local' localpath.

        :param remotepath: path to local file
        :param localpath: absolute path to remote file
        :param overwrite: if True overwrites localpath.
                               Default = False

        :raise OSError if 'remote' remotepath is not valid or not found
        :raise ValueError: if 'local' localpath is not valid
        :raise OSError: if unintentionally overwriting
        """
        overwrite = kwargs.get('overwrite', args[0] if args else True)
        if not localpath:
            raise ValueError('Input localpath to get function must be a non empty string')
        if not remotepath:
            raise OSError('Input remotepath to get function must be a non empty string')
        the_source = os.path.join(self.curdir, remotepath)
        if not os.path.exists(the_source):
            raise OSError('Source not found')
        if not os.path.isabs(localpath):
            raise ValueError('Destination must be an absolute path')
        if os.path.exists(localpath) and not overwrite:
            raise OSError('Destination already exists: not overwriting it')

        shutil.copyfile(the_source, localpath)

    def gettree(self, remotepath, localpath, *args, **kwargs):
        """Copies a folder recursively from 'remote' remotepath to
        'local' localpath.

        :param remotepath: path to local file
        :param localpath: absolute path to remote file
        :param dereference: follow symbolic links. Default = True
        :param overwrite: if True overwrites localpath. Default = False

        :raise OSError: if 'remote' remotepath is not valid
        :raise ValueError: if 'local' localpath is not valid
        :raise OSError: if unintentionally overwriting
        """
        dereference = kwargs.get('dereference', args[0] if args else True)
        overwrite = kwargs.get('overwrite', args[1] if len(args) > 1 else True)
        if not remotepath:
            raise OSError('Remotepath must be a non empty string')
        if not localpath:
            raise ValueError('Localpaths must be a non empty string')

        if not os.path.isabs(localpath):
            raise ValueError('Localpaths must be an absolute path')

        if not self.isdir(remotepath):
            raise OSError(f'Input remotepath is not a folder: {remotepath}')

        if os.path.exists(localpath) and not overwrite:
            raise OSError("Can't overwrite existing files")
        if os.path.isfile(localpath):
            raise OSError('Cannot copy a directory into a file')

        if os.path.isdir(localpath):
            localpath = os.path.join(localpath, os.path.split(remotepath)[1])

        the_source = os.path.join(self.curdir, remotepath)
        shutil.copytree(the_source, localpath, symlinks=not dereference, dirs_exist_ok=overwrite)

    # please refactor: issue #1780 on github

    def copy(self, remotesource, remotedestination, dereference=False, recursive=True):
        """Copies a file or a folder from 'remote' remotesource to 'remote' remotedestination.
        Automatically redirects to copyfile or copytree.

        :param remotesource: path to local file
        :param remotedestination: path to remote file
        :param dereference: follow symbolic links. Default = False
        :param recursive: if True copy directories recursively, otherwise only copy the specified file(s)
        :type recursive: bool

        :raise ValueError: if 'remote' remotesource or remotedestinationis not valid
        :raise OSError: if remotesource does not exist
        """
        if not remotesource:
            raise ValueError('Input remotesource to copy must be a non empty object')
        if not remotedestination:
            raise ValueError('Input remotedestination to copy must be a non empty object')
        if not self.has_magic(remotesource):
            if not os.path.exists(os.path.join(self.curdir, remotesource)):
                raise FileNotFoundError('Source not found')
        if self.normalize(remotesource) == self.normalize(remotedestination):
            raise ValueError('Cannot copy from itself to itself')

            # # by default, overwrite old files
        #        if not remotedestination .startswith('.'):
        #            if self.isfile(remotedestination) or self.isdir(remotedestination):
        #                self.rmtree(remotedestination)

        the_destination = os.path.join(self.curdir, remotedestination)

        if self.has_magic(remotesource):
            if self.has_magic(remotedestination):
                raise ValueError('Pathname patterns are not allowed in the remotedestination')

            to_copy_list = self.glob(remotesource)

            if len(to_copy_list) > 1:
                if not self.path_exists(remotedestination) or self.isfile(remotedestination):
                    raise OSError("Can't copy more than one file in the same remotedestination file")

            for source in to_copy_list:
                # If s is an absolute path, then the_s = s
                the_s = os.path.join(self.curdir, source)
                if self.isfile(source):
                    # With shutil, use the full path (the_s)
                    shutil.copy(the_s, the_destination)
                else:
                    # With self.copytree, the (possible) relative path is OK
                    self.copytree(source, remotedestination, dereference)

        else:
            # If s is an absolute path, then the_source = remotesource
            the_source = os.path.join(self.curdir, remotesource)
            if self.isfile(remotesource):
                # With shutil, use the full path (the_source)
                shutil.copy(the_source, the_destination)
            else:
                # With self.copytree, the (possible) relative path is OK
                self.copytree(remotesource, remotedestination, dereference)

    def copyfile(self, remotesource, remotedestination, dereference=False):
        """Copies a file from 'remote' remotesource to
        'remote' remotedestination.

        :param remotesource: path to local file
        :param remotedestination: path to remote file
        :param dereference: if True copy the contents of any symlinks found, otherwise copy the symlinks themselves
        :type dereference: bool
        :raise ValueError: if 'remote' remotesource or remotedestination is not valid
        :raise OSError: if remotesource does not exist
        """
        if not remotesource:
            raise ValueError('Input remotesource to copyfile must be a non empty object')
        if not remotedestination:
            raise ValueError('Input remotedestination to copyfile must be a non empty object')
        the_source = os.path.join(self.curdir, remotesource)
        the_destination = os.path.join(self.curdir, remotedestination)
        if not os.path.exists(the_source):
            raise OSError('Source not found')

        if not dereference and os.path.islink(remotesource):
            linkto = os.readlink(the_source)
            os.symlink(linkto, the_destination)
        else:
            shutil.copyfile(the_source, the_destination)

    def copytree(self, remotesource, remotedestination, dereference=False):
        """Copies a folder from 'remote' remotesource to
        'remote' remotedestination.

        :param remotesource: path to local file
        :param remotedestination: path to remote file
        :param dereference: follow symbolic links. Default = False

        :raise ValueError: if 'remote' remotesource or remotedestination is not valid
        :raise OSError: if remotesource does not exist
        """
        if not remotesource:
            raise ValueError('Input remotesource to copytree must be a non empty object')
        if not remotedestination:
            raise ValueError('Input remotedestination to copytree must be a non empty object')
        the_source = os.path.join(self.curdir, remotesource)
        the_destination = os.path.join(self.curdir, remotedestination)
        if not os.path.exists(the_source):
            raise OSError('Source not found')

        # Using the Ubuntu default behavior (different from Mac)
        if self.isdir(remotedestination):
            the_destination = os.path.join(the_destination, os.path.split(remotesource)[1])

        shutil.copytree(the_source, the_destination, symlinks=not dereference)

    def get_attribute(self, path):
        """Returns an object FileAttribute,
        as specified in aiida.transports.
        :param path: the path of the given file.
        """
        from aiida.transports.util import FileAttribute

        os_attr = os.lstat(os.path.join(self.curdir, path))
        aiida_attr = FileAttribute()
        # map the paramiko class into the aiida one
        # note that paramiko object contains more informations than the aiida
        for key in aiida_attr._valid_fields:
            aiida_attr[key] = getattr(os_attr, key)
        return aiida_attr

    def _local_listdir(self, path, pattern=None):
        """Act on the local folder, for the rest, same as listdir."""
        import re

        if not pattern:
            return os.listdir(path)

        if path.startswith('/'):  # always this is the case in the local plugin
            base_dir = path
        else:
            base_dir = os.path.join(os.getcwd(), path)

        filtered_list = glob.glob(os.path.join(base_dir, pattern))
        if not base_dir.endswith(os.sep):
            base_dir += os.sep
        return [re.sub(base_dir, '', i) for i in filtered_list]

    def listdir(self, path='.', pattern=None):
        """:return: a list containing the names of the entries in the directory.
        :param path: default ='.'
        :param pattern: if set, returns the list of files matching pattern.
                     Unix only. (Use to emulate ls * for example)
        """
        the_path = os.path.join(self.curdir, path).strip()
        if not pattern:
            try:
                return os.listdir(the_path)
            except OSError as err:
                exc = OSError(str(err))
                exc.errno = err.errno
                raise exc
        else:
            import re

            filtered_list = glob.glob(os.path.join(the_path, pattern))
            if not the_path.endswith('/'):
                the_path += '/'
            return [re.sub(the_path, '', i) for i in filtered_list]

    def remove(self, path):
        """Removes a file at position path."""
        os.remove(os.path.join(self.curdir, path))

    def isfile(self, path):
        """Checks if object at path is a file.
        Returns a boolean.
        """
        if not path:
            return False
        return os.path.isfile(os.path.join(self.curdir, path))

    @contextlib.contextmanager
    def _exec_command_internal(self, command, **kwargs):
        """Executes the specified command in bash login shell.

        Before the command is executed, changes directory to the current
        working directory as returned by self.getcwd().

        For executing commands and waiting for them to finish, use
        exec_command_wait.
        Otherwise, to end the process, use the proc.wait() method.

        The subprocess is set to have a different process group than the
        main process, so that it is shielded from signals sent to the parent.

        :param  command: the command to execute. The command is assumed to be
            already escaped using :py:func:`aiida.common.escaping.escape_for_bash`.

        :return: a tuple with (stdin, stdout, stderr, proc),
            where stdin, stdout and stderr behave as file-like objects,
            proc is the process object as returned by the
            subprocess.Popen() class.
        """
        from aiida.common.escaping import escape_for_bash

        # Note: The outer shell will eat one level of escaping, while
        # 'bash -l -c ...' will eat another. Thus, we need to escape again.
        bash_commmand = f'{self._bash_command_str}-c '

        command = bash_commmand + escape_for_bash(command)

        with subprocess.Popen(
            command,
            shell=True,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=self.getcwd(),
            start_new_session=True,
        ) as process:
            yield process

    def exec_command_wait_bytes(self, command, stdin=None, **kwargs):
        """Executes the specified command and waits for it to finish.

        :param command: the command to execute

        :return: a tuple with (return_value, stdout, stderr) where stdout and stderr
            are both bytes and the return_value is an int.
        """
        with self._exec_command_internal(command) as process:
            if stdin is not None:
                # Implicitly assume that the desired encoding is 'utf-8' if I receive a string.
                # Also, if I get a StringIO, I just read it all in memory and put it into a BytesIO.
                # Clearly not memory effective - in this case do not use a StringIO, but pass directly a BytesIO
                # that will be read line by line, if you have a huge stdin and care about memory usage.
                if isinstance(stdin, str):
                    filelike_stdin = io.BytesIO(stdin.encode('utf-8'))
                elif isinstance(stdin, bytes):
                    filelike_stdin = io.BytesIO(stdin)
                elif isinstance(stdin, io.TextIOBase):

                    def line_encoder(iterator, encoding='utf-8'):
                        """Encode the iterator item by item (i.e., line by line).

                        This only wraps iterating over it and not all other methods, but it's enough for its
                        use below.
                        """
                        for line in iterator:
                            yield line.encode(encoding)

                    filelike_stdin = line_encoder(stdin)
                elif isinstance(stdin, io.BufferedIOBase):
                    filelike_stdin = stdin
                else:
                    raise ValueError('You can only pass strings, bytes, BytesIO or StringIO objects')

                try:
                    for line in filelike_stdin:
                        process.stdin.write(line)  # the Popen.stdin/out/err are byte streams
                except AttributeError:
                    raise ValueError('stdin can only be either a string or a file-like object!')
            else:
                filelike_stdin = None

            process.stdin.flush()
            output_text, stderr_text = process.communicate()

            retval = process.returncode

        return retval, output_text, stderr_text

    def gotocomputer_command(self, remotedir):
        """Return a string to be run using os.system in order to connect
        via the transport to the remote directory.

        Expected behaviors:

        * A new bash session is opened
        * A reasonable error message is produced if the folder does not exist

        :param str remotedir: the full path of the remote directory
        """
        connect_string = self._gotocomputer_string(remotedir)
        cmd = f'bash -c {connect_string}'
        return cmd

    def rename(self, oldpath, newpath):
        """Rename a file or folder from oldpath to newpath.

        :param str oldpath: existing name of the file or folder
        :param str newpath: new name for the file or folder

        :raises OSError: if src/dst is not found
        :raises ValueError: if src/dst is not a valid string
        """
        if not oldpath:
            raise ValueError(f'Source {oldpath} is not a valid string')
        if not newpath:
            raise ValueError(f'Destination {newpath} is not a valid string')
        if not os.path.exists(oldpath):
            raise OSError(f'Source {oldpath} does not exist')
        if not os.path.exists(newpath):
            raise OSError(f'Destination {newpath} does not exist')

        shutil.move(oldpath, newpath)

    def symlink(self, remotesource, remotedestination):
        """Create a symbolic link between the remote source and the remote
        remotedestination

        :param remotesource: remote source. Can contain a pattern.
        :param remotedestination: remote destination
        """
        remotesource = os.path.normpath(remotesource)
        remotedestination = os.path.normpath(remotedestination)

        if self.has_magic(remotesource):
            if self.has_magic(remotedestination):
                # if there are patterns in dest, I don't know which name to assign
                raise ValueError('Remotedestination cannot have patterns')

            # find all files matching pattern
            for this_file in self.glob(remotesource):
                # create the name of the link: take the last part of the path
                this_remote_dest = os.path.split(this_file)[-1]

                os.symlink(os.path.join(this_file), os.path.join(self.curdir, remotedestination, this_remote_dest))
        else:
            try:
                os.symlink(remotesource, os.path.join(self.curdir, remotedestination))
            except OSError:
                raise OSError(f'!!: {remotesource}, {self.curdir}, {remotedestination}')

    def path_exists(self, path):
        """Check if path exists"""
        return os.path.exists(os.path.join(self.curdir, path))


CONFIGURE_LOCAL_CMD = transport_cli.create_configure_cmd('core.local')



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Abstract data plugin representing an executable code."""

from __future__ import annotations

import abc
import functools
import pathlib
import typing as t

from pydantic import BaseModel, field_validator

from aiida.cmdline.params.options.interactive import TemplateInteractiveOption
from aiida.common import exceptions
from aiida.common.folders import Folder
from aiida.common.lang import type_check
from aiida.common.pydantic import MetadataField
from aiida.orm import Computer
from aiida.plugins import CalculationFactory

from ..data import Data

if t.TYPE_CHECKING:
    from aiida.engine import ProcessBuilder

__all__ = ('AbstractCode',)


class AbstractCode(Data, metaclass=abc.ABCMeta):
    """Abstract data plugin representing an executable code."""

    # Should become ``default_calc_job_plugin`` once ``Code`` is dropped in ``aiida-core==3.0``
    _KEY_ATTRIBUTE_DEFAULT_CALC_JOB_PLUGIN: str = 'input_plugin'
    _KEY_ATTRIBUTE_APPEND_TEXT: str = 'append_text'
    _KEY_ATTRIBUTE_PREPEND_TEXT: str = 'prepend_text'
    _KEY_ATTRIBUTE_USE_DOUBLE_QUOTES: str = 'use_double_quotes'
    _KEY_ATTRIBUTE_WITH_MPI: str = 'with_mpi'
    _KEY_ATTRIBUTE_WRAP_CMDLINE_PARAMS: str = 'wrap_cmdline_params'
    _KEY_EXTRA_IS_HIDDEN: str = 'hidden'  # Should become ``is_hidden`` once ``Code`` is dropped

    class Model(BaseModel, defer_build=True):
        """Model describing required information to create an instance."""

        label: str = MetadataField(
            ...,
            title='Label',
            description='A unique label to identify the code by.',
            short_name='-L',
        )
        description: str = MetadataField(
            '',
            title='Description',
            description='Human-readable description, ideally including version and compilation environment.',
            short_name='-D',
        )
        default_calc_job_plugin: t.Optional[str] = MetadataField(
            None,
            title='Default `CalcJob` plugin',
            description='Entry point name of the default plugin (as listed in `verdi plugin list aiida.calculations`).',
            short_name='-P',
        )
        use_double_quotes: bool = MetadataField(
            False,
            title='Escape using double quotes',
            description='Whether the executable and arguments of the code in the submission script should be escaped '
            'with single or double quotes.',
        )
        with_mpi: t.Optional[bool] = MetadataField(
            None,
            title='Run with MPI',
            description='Whether the executable should be run as an MPI program. This option can be left unspecified '
            'in which case `None` will be set and it is left up to the calculation job plugin or inputs '
            'whether to run with MPI.',
        )
        prepend_text: str = MetadataField(
            '',
            title='Prepend script',
            description='Bash commands that should be prepended to the run line in all submit scripts for this code.',
            option_cls=functools.partial(
                TemplateInteractiveOption,
                extension='.bash',
                header='PREPEND_TEXT: if there is any bash commands that should be prepended to the executable call '
                'in all submit scripts for this code, type that between the equal signs below and save the file.',
                footer='All lines that start with `#=`: will be ignored.',
            ),
        )
        append_text: str = MetadataField(
            '',
            title='Append script',
            description='Bash commands that should be appended to the run line in all submit scripts for this code.',
            option_cls=functools.partial(
                TemplateInteractiveOption,
                extension='.bash',
                header='APPEND_TEXT: if there is any bash commands that should be appended to the executable call '
                'in all submit scripts for this code, type that between the equal signs below and save the file.',
                footer='All lines that start with `#=`: will be ignored.',
            ),
        )

        @field_validator('label')
        @classmethod
        def validate_label_uniqueness(cls, value: str) -> str:
            """Validate that the label does not already exist."""
            from aiida.orm import load_code

            try:
                load_code(value)
            except exceptions.NotExistent:
                return value
            except exceptions.MultipleObjectsError as exception:
                raise ValueError(f'Multiple codes with the label `{value}` already exist.') from exception
            else:
                raise ValueError(f'A code with the label `{value}` already exists.')

    def __init__(
        self,
        default_calc_job_plugin: str | None = None,
        append_text: str = '',
        prepend_text: str = '',
        use_double_quotes: bool = False,
        with_mpi: bool | None = None,
        is_hidden: bool = False,
        wrap_cmdline_params: bool = False,
        **kwargs,
    ):
        """Construct a new instance.

        :param default_calc_job_plugin: The entry point name of the default ``CalcJob`` plugin to use.
        :param append_text: The text that should be appended to the run line in the job script.
        :param prepend_text: The text that should be prepended to the run line in the job script.
        :param use_double_quotes: Whether the command line invocation of this code should be escaped with double quotes.
        :param with_mpi: Whether the command should be run as an MPI program.
        :param wrap_cmdline_params: Whether to wrap the executable and all its command line parameters into quotes to
            form a single string. This is required to enable support for Docker with the ``ContainerizedCode``.
        :param is_hidden: Whether the code is hidden.
        """
        super().__init__(**kwargs)
        self.default_calc_job_plugin = default_calc_job_plugin
        self.append_text = append_text
        self.prepend_text = prepend_text
        self.use_double_quotes = use_double_quotes
        self.with_mpi = with_mpi
        self.wrap_cmdline_params = wrap_cmdline_params
        self.is_hidden = is_hidden

    @abc.abstractmethod
    def can_run_on_computer(self, computer: Computer) -> bool:
        """Return whether the code can run on a given computer.

        :param computer: The computer.
        :return: ``True`` if the code can run on ``computer``, ``False`` otherwise.
        """

    @abc.abstractmethod
    def get_executable(self) -> pathlib.PurePosixPath:
        """Return the executable that the submission script should execute to run the code.

        :return: The executable to be called in the submission script.
        """

    def get_executable_cmdline_params(self, cmdline_params: list[str] | None = None) -> list:
        """Return the list of executable with its command line parameters.

        :param cmdline_params: List of command line parameters provided by the ``CalcJob`` plugin.
        :return: List of the executable followed by its command line parameters.
        """
        return [str(self.get_executable())] + (cmdline_params or [])

    def get_prepend_cmdline_params(
        self, mpi_args: list[str] | None = None, extra_mpirun_params: list[str] | None = None
    ) -> list[str]:
        """Return List of command line parameters to be prepended to the executable in submission line.
        These command line parameters are typically parameters related to MPI invocations.

        :param mpi_args: List of MPI parameters provided by the ``Computer.get_mpirun_command`` method.
        :param extra_mpiruns_params: List of MPI parameters provided by the ``metadata.options.extra_mpirun_params``
            input of the ``CalcJob``.
        :return: List of command line parameters to be prepended to the executable in submission line.
        """
        return (mpi_args or []) + (extra_mpirun_params or [])

    def validate_working_directory(self, folder: Folder):
        """Validate content of the working directory created by the :class:`~aiida.engine.CalcJob` plugin.

        This method will be called by :meth:`~aiida.engine.processes.calcjobs.calcjob.CalcJob.presubmit` when a new
        calculation job is launched, passing the :class:`~aiida.common.folders.Folder` that was used by the plugin used
        for the calculation to create the input files for the working directory. This method can be overridden by
        implementations of the ``AbstractCode`` class that need to validate the contents of that folder.

        :param folder: A sandbox folder that the ``CalcJob`` plugin wrote input files to that will be copied to the
            working directory for the corresponding calculation job instance.
        :raises PluginInternalError: If the content of the sandbox folder is not valid.
        """

    @property
    @abc.abstractmethod
    def full_label(self) -> str:
        """Return the full label of this code.

        The full label can be just the label itself but it can be something else. However, it at the very least has to
        include the label of the code.

        :return: The full label of the code.
        """

    @property
    def label(self) -> str:
        """Return the label.

        :return: The label.
        """
        return self.backend_entity.label

    @label.setter
    def label(self, value: str) -> None:
        """Set the label.

        The label cannot contain any ``@`` symbols.

        :param value: The new label.
        :raises ValueError: If the label contains invalid characters.
        """
        type_check(value, str)

        if '@' in value:
            raise ValueError('The label contains a `@` symbol, which is not allowed.')

        self.backend_entity.label = value

    @property
    def default_calc_job_plugin(self) -> str | None:
        """Return the optional default ``CalcJob`` plugin.

        :return: The entry point name of the default ``CalcJob`` plugin to use.
        """
        return self.base.attributes.get(self._KEY_ATTRIBUTE_DEFAULT_CALC_JOB_PLUGIN, None)

    @default_calc_job_plugin.setter
    def default_calc_job_plugin(self, value: str | None) -> None:
        """Set the default ``CalcJob`` plugin.

        :param value: The entry point name of the default ``CalcJob`` plugin to use.
        """
        type_check(value, str, allow_none=True)
        self.base.attributes.set(self._KEY_ATTRIBUTE_DEFAULT_CALC_JOB_PLUGIN, value)

    @property
    def append_text(self) -> str:
        """Return the text that should be appended to the run line in the job script.

        :return: The text that should be appended to the run line in the job script.
        """
        return self.base.attributes.get(self._KEY_ATTRIBUTE_APPEND_TEXT, '')

    @append_text.setter
    def append_text(self, value: str) -> None:
        """Set the text that should be appended to the run line in the job script.

        :param value: The text that should be appended to the run line in the job script.
        """
        type_check(value, str, allow_none=True)
        self.base.attributes.set(self._KEY_ATTRIBUTE_APPEND_TEXT, value)

    @property
    def prepend_text(self) -> str:
        """Return the text that should be prepended to the run line in the job script.

        :return: The text that should be prepended to the run line in the job script.
        """
        return self.base.attributes.get(self._KEY_ATTRIBUTE_PREPEND_TEXT, '')

    @prepend_text.setter
    def prepend_text(self, value: str) -> None:
        """Set the text that should be prepended to the run line in the job script.

        :param value: The text that should be prepended to the run line in the job script.
        """
        type_check(value, str, allow_none=True)
        self.base.attributes.set(self._KEY_ATTRIBUTE_PREPEND_TEXT, value)

    @property
    def use_double_quotes(self) -> bool:
        """Return whether the command line invocation of this code should be escaped with double quotes.

        :return: ``True`` if to escape with double quotes, ``False`` otherwise.
        """
        return self.base.attributes.get(self._KEY_ATTRIBUTE_USE_DOUBLE_QUOTES, False)

    @use_double_quotes.setter
    def use_double_quotes(self, value: bool) -> None:
        """Set whether the command line invocation of this code should be escaped with double quotes.

        :param value: ``True`` if to escape with double quotes, ``False`` otherwise.
        """
        type_check(value, bool)
        self.base.attributes.set(self._KEY_ATTRIBUTE_USE_DOUBLE_QUOTES, value)

    @property
    def with_mpi(self) -> bool | None:
        """Return whether the command should be run as an MPI program.

        :return: ``True`` if the code should be run as an MPI program, ``False`` if it shouldn't, ``None`` if unknown.
        """
        return self.base.attributes.get(self._KEY_ATTRIBUTE_WITH_MPI, None)

    @with_mpi.setter
    def with_mpi(self, value: bool | None) -> None:
        """Set whether the command should be run as an MPI program.

        :param value: ``True`` if the code should be run as an MPI program, ``False`` if it shouldn't, ``None`` if
            unknown.
        """
        type_check(value, bool, allow_none=True)
        self.base.attributes.set(self._KEY_ATTRIBUTE_WITH_MPI, value)

    @property
    def wrap_cmdline_params(self) -> bool:
        """Return whether all command line parameters should be wrapped with double quotes to form a single argument.

        ..note:: This is required to support certain containerization technologies, such as Docker.

        :return: ``True`` if command line parameters should be wrapped, ``False`` otherwise.
        """
        return self.base.attributes.get(self._KEY_ATTRIBUTE_WRAP_CMDLINE_PARAMS, False)

    @wrap_cmdline_params.setter
    def wrap_cmdline_params(self, value: bool) -> None:
        """Set whether all command line parameters should be wrapped with double quotes to form a single argument.

        :param value: ``True`` if command line parameters should be wrapped, ``False`` otherwise.
        """
        type_check(value, bool)
        self.base.attributes.set(self._KEY_ATTRIBUTE_WRAP_CMDLINE_PARAMS, value)

    @property
    def is_hidden(self) -> bool:
        """Return whether the code is hidden.

        :return: ``True`` if the code is hidden, ``False`` otherwise, which is also the default.
        """
        return self.base.extras.get(self._KEY_EXTRA_IS_HIDDEN, False)

    @is_hidden.setter
    def is_hidden(self, value: bool) -> None:
        """Define whether the code is hidden or not.

        :param value: ``True`` if the code should be hidden, ``False`` otherwise.
        """
        type_check(value, bool)
        self.base.extras.set(self._KEY_EXTRA_IS_HIDDEN, value)

    def get_builder(self) -> 'ProcessBuilder':
        """Create and return a new ``ProcessBuilder`` for the ``CalcJob`` class of the plugin configured for this code.

        The configured calculation plugin class is defined by the ``default_calc_job_plugin`` property.

        .. note:: it also sets the ``builder.code`` value.

        :return: a ``ProcessBuilder`` instance with the ``code`` input already populated with ourselves
        :raise aiida.common.EntryPointError: if the specified plugin does not exist.
        :raise ValueError: if no default plugin was specified.
        """
        entry_point = self.default_calc_job_plugin

        if entry_point is None:
            raise ValueError('No default calculation input plugin specified for this code')

        try:
            process_class = CalculationFactory(entry_point)
        except exceptions.EntryPointError:
            raise exceptions.EntryPointError(f'The calculation entry point `{entry_point}` could not be loaded')

        builder = process_class.get_builder()  # type: ignore[union-attr]
        builder.code = self

        return builder



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Backend group module"""

import abc
from typing import TYPE_CHECKING, List, Optional, Protocol, Sequence, Union

from .entities import BackendCollection, BackendEntity, BackendEntityExtrasMixin
from .nodes import BackendNode

if TYPE_CHECKING:
    from .users import BackendUser

__all__ = ('BackendGroup', 'BackendGroupCollection')


class NodeIterator(Protocol):
    """Protocol for iterating over nodes in a group"""

    def __iter__(self) -> 'NodeIterator':
        """Return an iterator over the nodes in the group."""

    def __next__(self) -> BackendNode:
        """Return the next node in the group."""

    def __getitem__(self, value: Union[int, slice]) -> Union[BackendNode, List[BackendNode]]:
        """Index node(s) from the group."""

    def __len__(self) -> int:
        """Return the number of nodes in the group."""


class BackendGroup(BackendEntity, BackendEntityExtrasMixin):
    """Backend implementation for the `Group` ORM class.

    A group is a collection of nodes.
    """

    @property
    @abc.abstractmethod
    def label(self) -> str:
        """Return the name of the group as a string."""

    @label.setter
    @abc.abstractmethod
    def label(self, name: str) -> None:
        """Attempt to change the name of the group instance. If the group is already stored
        and the another group of the same type already exists with the desired name, a
        UniquenessError will be raised

        :param name: the new group name
        :raises aiida.common.UniquenessError: if another group of same type and name already exists
        """

    @property
    @abc.abstractmethod
    def description(self) -> Optional[str]:
        """Return the description of the group as a string."""

    @description.setter
    @abc.abstractmethod
    def description(self, value: Optional[str]):
        """Return the description of the group as a string."""

    @property
    @abc.abstractmethod
    def type_string(self) -> str:
        """Return the string defining the type of the group."""

    @property
    @abc.abstractmethod
    def user(self) -> 'BackendUser':
        """Return a backend user object, representing the user associated to this group."""

    @user.setter
    @abc.abstractmethod
    def user(self, user: 'BackendUser') -> None:
        """Set the user of this group."""

    @property
    @abc.abstractmethod
    def uuid(self) -> str:
        """Return the UUID of the group."""

    @property
    @abc.abstractmethod
    def nodes(self) -> NodeIterator:
        """Return a generator/iterator that iterates over all nodes and returns
        the respective AiiDA subclasses of Node, and also allows to ask for
        the number of nodes in the group using len().
        """

    @abc.abstractmethod
    def count(self) -> int:
        """Return the number of entities in this group.

        :return: integer number of entities contained within the group
        """

    @abc.abstractmethod
    def clear(self) -> None:
        """Remove all the nodes from this group."""

    def add_nodes(self, nodes: Sequence[BackendNode], **kwargs):
        """Add a set of nodes to the group.

        :note: all the nodes *and* the group itself have to be stored.

        :param nodes: a list of `BackendNode` instances to be added to this group
        """
        if not self.is_stored:
            raise ValueError('group has to be stored before nodes can be added')

        if not isinstance(nodes, (list, tuple)):
            raise TypeError('nodes has to be a list or tuple')

        if any(not isinstance(node, BackendNode) for node in nodes):
            raise TypeError(f'nodes have to be of type {BackendNode}')

    def remove_nodes(self, nodes: Sequence[BackendNode]) -> None:
        """Remove a set of nodes from the group.

        :note: all the nodes *and* the group itself have to be stored.

        :param nodes: a list of `BackendNode` instances to be removed from this group
        """
        if not self.is_stored:
            raise ValueError('group has to be stored before nodes can be removed')

        if not isinstance(nodes, (list, tuple)):
            raise TypeError('nodes has to be a list or tuple')

        if any(not isinstance(node, BackendNode) for node in nodes):
            raise TypeError(f'nodes have to be of type {BackendNode}')

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}: {self!s}>'

    def __str__(self) -> str:
        if self.type_string:
            return f'"{self.label}" [type {self.type_string}], of user {self.user.email}'

        return f'"{self.label}" [user-defined], of user {self.user.email}'


class BackendGroupCollection(BackendCollection[BackendGroup]):
    """The collection of Group entries."""

    ENTITY_CLASS = BackendGroup

    @abc.abstractmethod
    def delete(self, id: int) -> None:
        """Delete a group with the given id

        :param id: the id of the group to delete
        """



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Update all link labels with the value `_return`

This is the legacy default single link label.
The old process functions used to use `_return` as the default link label, however, since labels that start or end with
and underscore are illegal because they are used for namespacing.

This is identical to migration django_0043

Revision ID: 118349c10896
Revises: 91b573400be5
Create Date: 2019-11-21 09:43:45.006053

"""

from alembic import op
from sqlalchemy.sql import text

# revision identifiers, used by Alembic.
revision = '118349c10896'
down_revision = '91b573400be5'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    conn = op.get_bind()

    # The old process functions used to use `_return` as the default link label, however, since labels that start or end
    # with and underscore are illegal.
    statement = text(
        """
        UPDATE db_dblink SET label='result' WHERE label = '_return';
    """
    )
    conn.execute(statement)


def downgrade():
    """Migrations for the downgrade."""
    raise NotImplementedError('Downgrade of 118349c10896.')
